# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import warnings
from pathlib import Path

import neurom as nm
import numpy as np
import pandas as pd
from neurom.apps import morph_stats as ms
from neurom.exceptions import ConfigError
from neurom.features import NEURITEFEATURES, NEURONFEATURES

import pytest
from numpy.testing import assert_array_equal, assert_almost_equal
from pandas.testing import assert_frame_equal

DATA_PATH = Path(__file__).parent.parent / 'data'
SWC_PATH = DATA_PATH / 'swc'
REF_CONFIG = {
    'neurite': {
        'section_lengths': ['max', 'total'],
        'section_volumes': ['total'],
        'section_branch_orders': ['max'],
        'segment_midpoints': ['max'],
        'max_radial_distance': ['mean'],
    },
    'neurite_type': ['AXON', 'APICAL_DENDRITE', 'BASAL_DENDRITE', 'ALL'],
    'neuron': {
        'soma_radii': ['mean'],
        'max_radial_distance': ['mean'],
    }
}

REF_OUT = {
    'neuron': {
        'mean_soma_radius': 0.13065629648763766,
        'mean_max_radial_distance': 99.5894610648815,
    },
    'axon': {
        'total_section_length': 207.87975220908129,
        'max_section_length': 11.018460736176685,
        'max_section_branch_order': 10,
        'total_section_volume': 276.73857657289523,
        'max_segment_midpoint_0': 0.0,
        'max_segment_midpoint_1': 0.0,
        'max_segment_midpoint_2': 49.520305964149998,
        'mean_max_radial_distance': 82.44254511788921,
    },
    'all': {
        'total_section_length': 840.68521442251949,
        'max_section_length': 11.758281556059444,
        'max_section_branch_order': 10,
        'total_section_volume': 1104.9077419665782,
        'max_segment_midpoint_0': 64.401674984050004,
        'max_segment_midpoint_1': 48.48197694465,
        'max_segment_midpoint_2': 53.750947521650005,
        'mean_max_radial_distance': 99.5894610648815,
    },
    'apical_dendrite': {
        'total_section_length': 214.37304577550353,
        'max_section_length': 11.758281556059444,
        'max_section_branch_order': 10,
        'total_section_volume': 271.9412385728449,
        'max_segment_midpoint_0': 64.401674984050004,
        'max_segment_midpoint_1': 0.0,
        'max_segment_midpoint_2': 53.750947521650005,
        'mean_max_radial_distance': 99.5894610648815,
    },
    'basal_dendrite': {
        'total_section_length': 418.43241643793476,
        'max_section_length': 11.652508126101711,
        'max_section_branch_order': 10,
        'total_section_volume': 556.22792682083821,
        'max_segment_midpoint_0': 64.007872333250006,
        'max_segment_midpoint_1': 48.48197694465,
        'max_segment_midpoint_2': 51.575580778049996,
        'mean_max_radial_distance': 94.43342438865741,
    },
}


def test_name_correction():
    assert ms._stat_name('foo', 'raw') == 'foo'
    assert ms._stat_name('foos', 'raw') == 'foo'
    assert ms._stat_name('foos', 'bar') == 'bar_foo'
    assert ms._stat_name('foos', 'total') == 'total_foo'
    assert ms._stat_name('soma_radii', 'total') == 'total_soma_radius'
    assert ms._stat_name('soma_radii', 'raw') == 'soma_radius'


def test_eval_stats_raw_returns_list():
    assert ms.eval_stats(np.array([1, 2, 3, 4]), 'raw') == [1, 2, 3, 4]


def test_eval_stats_empty_input_returns_none():
    assert ms.eval_stats([], 'min') is None


def test_eval_stats_total_returns_sum():
    assert ms.eval_stats(np.array([1, 2, 3, 4]), 'total') == 10


def test_eval_stats_on_empty_stat():
    assert ms.eval_stats(np.array([]), 'mean') == None
    assert ms.eval_stats(np.array([]), 'std') == None
    assert ms.eval_stats(np.array([]), 'median') == None
    assert ms.eval_stats(np.array([]), 'min') == None
    assert ms.eval_stats(np.array([]), 'max') == None

    assert ms.eval_stats(np.array([]), 'raw') == []
    assert ms.eval_stats(np.array([]), 'total') == 0.0


def test_eval_stats_applies_numpy_function():
    modes = ('min', 'max', 'mean', 'median', 'std')

    ref_array = np.arange(1, 10)

    for m in modes:
        assert (ms.eval_stats(ref_array, m) == getattr(np, m)(ref_array))


def test_extract_stats_single_neuron():
    nrn = nm.load_neuron(SWC_PATH / 'Neuron.swc')
    res = ms.extract_stats(nrn, REF_CONFIG)
    assert set(res.keys()) == set(REF_OUT.keys())
    for k in ('neuron', 'all', 'axon', 'basal_dendrite', 'apical_dendrite'):
        assert set(res[k].keys()) == set(REF_OUT[k].keys())
        for kk in res[k].keys():
            assert_almost_equal(res[k][kk], REF_OUT[k][kk], decimal=4)


def test_extract_stats_scalar_feature():
    nrn = nm.load_neuron(DATA_PATH / 'neurolucida' / 'bio_neuron-000.asc')
    config = {
        'neurite_type': ['ALL'],
        'neurite': {
            'n_forking_points': ['max'],
        },
        'neuron': {
            'soma_volume': ['total'],
        }
    }
    res = ms.extract_stats(nrn, config)
    assert res == {'all': {'max_n_forking_point': 277},
                       'neuron': {'total_soma_volume': 1424.4383771584492}}


def test_extract_dataframe():
    # Vanilla test
    nrns = nm.load_neurons([SWC_PATH / name
                            for name in ['Neuron.swc', 'simple.swc']])
    actual = ms.extract_dataframe(nrns, REF_CONFIG)
    expected = pd.read_csv(Path(DATA_PATH, 'extracted-stats.csv'), header=[0, 1], index_col=0)
    assert_frame_equal(actual, expected)

    # Test with a single neuron in the population
    nrns = nm.load_neurons(Path(SWC_PATH, 'Neuron.swc'))
    actual = ms.extract_dataframe(nrns, REF_CONFIG)
    assert_frame_equal(actual, expected.iloc[[0]], check_dtype=False)

    # Test with a config without the 'neuron' key
    nrns = nm.load_neurons([Path(SWC_PATH, name)
                            for name in ['Neuron.swc', 'simple.swc']])
    config = {'neurite': {'section_lengths': ['total']},
              'neurite_type': ['AXON', 'APICAL_DENDRITE', 'BASAL_DENDRITE', 'ALL']}
    actual = ms.extract_dataframe(nrns, config)
    idx = pd.IndexSlice
    expected = expected.loc[:, idx[:, ['name', 'total_section_length']]]
    assert_frame_equal(actual, expected)

    # Test with a FstNeuron argument
    nrn = nm.load_neuron(Path(SWC_PATH, 'Neuron.swc'))
    actual = ms.extract_dataframe(nrn, config)
    assert_frame_equal(actual, expected.iloc[[0]], check_dtype=False)

    # Test with a List[FstNeuron] argument
    nrns = [nm.load_neuron(Path(SWC_PATH, name))
            for name in ['Neuron.swc', 'simple.swc']]
    actual = ms.extract_dataframe(nrns, config)
    assert_frame_equal(actual, expected)

    # Test with a List[Path] argument
    nrns = [Path(SWC_PATH, name) for name in ['Neuron.swc', 'simple.swc']]
    actual = ms.extract_dataframe(nrns, config)
    assert_frame_equal(actual, expected)

    # Test without any neurite_type keys, it should pick the defaults
    config = {'neurite': {'total_length_per_neurite': ['total']}}
    actual = ms.extract_dataframe(nrns, config)
    expected_columns = pd.MultiIndex.from_tuples(
        [('property', 'name'),
         ('axon', 'total_total_length_per_neurite'),
         ('basal_dendrite', 'total_total_length_per_neurite'),
         ('apical_dendrite', 'total_total_length_per_neurite'),
         ('all', 'total_total_length_per_neurite')])
    expected = pd.DataFrame(
        columns=expected_columns,
        data=[['Neuron', 207.87975221, 418.43241644, 214.37304578, 840.68521442],
              ['simple', 15.,          16.,           0.,          31., ]])
    assert_frame_equal(actual, expected)


def test_extract_dataframe_multiproc():
    # FIXME: Cannot use Neuron objects in the extract_dataframe ctor right now
    # because of "TypeError: can't pickle Neuron objects"
    # nrns = nm.load_neurons([Path(SWC_PATH, name)
    #                         for name in ['Neuron.swc', 'simple.swc']])
    nrns = [Path(SWC_PATH, name)
            for name in ['Neuron.swc', 'simple.swc']]
    with warnings.catch_warnings(record=True) as w:
        actual = ms.extract_dataframe(nrns, REF_CONFIG, n_workers=2)
    expected = pd.read_csv(Path(DATA_PATH, 'extracted-stats.csv'), index_col=0, header=[0, 1])

    assert_frame_equal(actual, expected)

    with warnings.catch_warnings(record=True) as w:
        actual = ms.extract_dataframe(nrns, REF_CONFIG, n_workers=os.cpu_count() + 1)
        assert len(w) == 1, "Warning not emitted"
    assert_frame_equal(actual, expected)


def test_get_header():
    fake_results = {'fake_name0': REF_OUT,
                    'fake_name1': REF_OUT,
                    'fake_name2': REF_OUT,
                    }
    header = ms.get_header(fake_results)
    assert 1 + 2 + 4 * (4 + 4) == len(header)  # name + everything in REF_OUT
    assert 'name' in header
    assert 'neuron:mean_soma_radius' in header


def test_generate_flattened_dict():
    fake_results = {'fake_name0': REF_OUT,
                    'fake_name1': REF_OUT,
                    'fake_name2': REF_OUT,
                    }
    header = ms.get_header(fake_results)
    rows = list(ms.generate_flattened_dict(header, fake_results))
    assert 3 == len(rows)  # one for fake_name[0-2]
    assert 1 + 2 + 4 * (4 + 4) == len(rows[0])  # name + everything in REF_OUT


def test_full_config():
    config = ms.full_config()
    assert set(config.keys()) == {'neurite', 'neuron', 'neurite_type'}

    assert set(config['neurite'].keys()) == set(NEURITEFEATURES.keys())
    assert set(config['neuron'].keys()) == set(NEURONFEATURES.keys())


def test_sanitize_config():

    with pytest.raises(ConfigError):
        ms.sanitize_config({'neurite': []})

    new_config = ms.sanitize_config({})  # empty
    assert 2 == len(new_config)  # neurite & neuron created

    full_config = {
        'neurite': {
            'section_lengths': ['max', 'total'],
            'section_volumes': ['total'],
            'section_branch_orders': ['max']
        },
        'neurite_type': ['AXON', 'APICAL_DENDRITE', 'BASAL_DENDRITE', 'ALL'],
        'neuron': {
            'soma_radii': ['mean']
        }
    }
    new_config = ms.sanitize_config(full_config)
    assert 3 == len(new_config)  # neurite, neurite_type & neuron


def test_multidimensional_features():
    """Features should be split into sub-features when they
    are multidimensional.


    This should be the case even when the feature is `None` or `[]`
    The following neuron has no axon but the axon feature segment_midpoints for
    the axon should still be made of 3 values (X, Y and Z)

    Cf: https://github.com/BlueBrain/NeuroM/issues/859
    """
    neuron = nm.load_neuron(Path(SWC_PATH, 'no-axon.swc'))

    config = {'neurite': {'segment_midpoints': ['max']},
              'neurite_type': ['AXON']}
    actual = ms.extract_dataframe(neuron, config)
    assert_array_equal(actual['axon'][['max_segment_midpoint_0',
                                       'max_segment_midpoint_1',
                                       'max_segment_midpoint_2']].values,
                       [[None, None, None]])

    config = {'neurite': {'partition_pairs': ['max']}}
    actual = ms.extract_dataframe(neuron, config)
    assert_array_equal(actual['axon'][['max_partition_pair_0',
                                       'max_partition_pair_1']].values,
                       [[None, None]])



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from pathlib import Path
import tempfile

import numpy as np
from neurom.view.common import (plt, figure_naming, get_figure, save_plot, plot_style,
                                plot_title, plot_labels, plot_legend, update_plot_limits, plot_ticks,
                                plot_sphere, plot_cylinder)
import pytest


def test_figure_naming():
    pretitle, posttitle, prefile, postfile = figure_naming(pretitle='Test', prefile="", postfile=3)
    assert pretitle == 'Test -- '
    assert posttitle == ""
    assert prefile == ""
    assert postfile == "_3"

    pretitle, posttitle, prefile, postfile = figure_naming(pretitle='', posttitle="Test", prefile="test", postfile="")
    assert pretitle == ""
    assert posttitle == " -- Test"
    assert prefile == "test_"
    assert postfile == ""


def test_get_figure():
    fig_old = plt.figure()
    fig, ax = get_figure(new_fig=False)
    assert fig == fig_old
    assert ax.get_subplotspec().colspan.start == 0
    assert ax.get_subplotspec().rowspan.start == 0

    fig1, ax1 = get_figure(new_fig=True, subplot=224)
    assert fig1 != fig_old
    assert ax1.get_subplotspec().colspan.start == 1
    assert ax1.get_subplotspec().rowspan.start == 1

    fig2, ax2 = get_figure(new_fig=True, subplot=[1, 1, 1])
    assert ax2.get_subplotspec().colspan.start == 0
    assert ax2.get_subplotspec().rowspan.start == 0
    plt.close('all')

    fig = plt.figure()
    ax = fig.add_subplot(111)
    fig2, ax2 = get_figure(new_fig=False)
    assert fig2 == plt.gcf()
    assert ax2 == plt.gca()
    plt.close('all')


def test_save_plot():
    with tempfile.TemporaryDirectory() as folder:
        fig_old = plt.figure()
        output_path = Path(folder, 'subdir')
        fig = save_plot(fig_old, output_path=output_path)
        assert Path(output_path, 'Figure.png').is_file()
        plt.close('all')


def test_plot_title_default(get_fig_2d):
    fig, ax = get_fig_2d
    plot_title(ax)
    assert ax.get_title() == 'Figure'


def test_plot_title(get_fig_2d):
    fig, ax = get_fig_2d
    plot_title(ax, title='Test')
    assert ax.get_title() == 'Test'


def test_plot_labels(get_fig_2d, get_fig_3d):
    fig, ax = get_fig_2d
    plot_labels(ax)
    assert ax.get_xlabel() == 'X'
    assert ax.get_ylabel() == 'Y'

    fig, ax = get_fig_2d
    plot_labels(ax, xlabel='T', ylabel='R')
    assert ax.get_xlabel() == 'T'
    assert ax.get_ylabel() == 'R'

    fig0, ax0 = get_fig_3d
    plot_labels(ax0)
    assert ax0.get_zlabel() == 'Z'

    fig0, ax0 = get_fig_3d
    plot_labels(ax0, zlabel='T')
    assert ax0.get_zlabel() == 'T'


def test_plot_legend(get_fig_2d):
    fig, ax = get_fig_2d
    plot_legend(ax)
    legend = ax.get_legend()
    assert legend is None

    fig, ax = get_fig_2d
    ax.plot([1, 2, 3], [1, 2, 3], label='line 1')
    plot_legend(ax, no_legend=False)
    legend = ax.get_legend()
    assert legend.get_texts()[0].get_text() == 'line 1'


def test_plot_limits(get_fig_2d, get_fig_3d):
    fig, ax = get_fig_2d
    with pytest.raises(AssertionError):
        update_plot_limits(ax, white_space=0)

    fig, ax = get_fig_2d
    ax.dataLim.update_from_data_xy(((0, -100), (100, 0)))

    update_plot_limits(ax, white_space=0)
    assert ax.get_xlim() == (0, 100)
    assert ax.get_ylim() == (-100, 0)

    fig0, ax0 = get_fig_3d
    update_plot_limits(ax0, white_space=0)
    zlim0 = ax0.get_zlim()
    assert np.allclose(ax0.get_zlim(), zlim0)


def test_plot_ticks(get_fig_2d, get_fig_3d):
    fig, ax = get_fig_2d
    plot_ticks(ax)
    assert len(ax.get_xticks())
    assert len(ax.get_yticks())

    fig, ax = get_fig_2d
    plot_ticks(ax, xticks=[], yticks=[])
    assert len(ax.get_xticks()) == 0
    assert len(ax.get_yticks()) == 0

    fig, ax = get_fig_2d
    plot_ticks(ax, xticks=np.arange(3), yticks=np.arange(4))
    assert len(ax.get_xticks()) == 3
    assert len(ax.get_yticks()) == 4

    fig0, ax0 = get_fig_3d
    plot_ticks(ax0)
    assert len(ax0.get_zticks())

    fig0, ax0 = get_fig_3d
    plot_ticks(ax0, zticks=[])
    assert len(ax0.get_zticks()) == 0

    fig0, ax0 = get_fig_3d
    plot_ticks(ax0, zticks=np.arange(3))
    assert len(ax0.get_zticks()) == 3


def test_plot_style(get_fig_2d):
    fig, ax = get_fig_2d
    ax.dataLim.update_from_data_xy(((0, -100), (100, 0)))

    plot_style(fig, ax)

    assert ax.get_title() == 'Figure'
    assert ax.get_xlabel() == 'X'
    assert ax.get_ylabel() == 'Y'

    fig, ax = get_fig_2d
    ax.dataLim.update_from_data_xy(((0, -100), (100, 0)))

    plot_style(fig, ax, no_axes=True)

    assert not ax.get_frame_on()
    assert not ax.xaxis.get_visible()
    assert not ax.yaxis.get_visible()

    fig, ax = get_fig_2d
    ax.dataLim.update_from_data_xy(((0, -100), (100, 0)))
    plot_style(fig, ax, tight=True)
    assert fig.get_tight_layout()


def test_plot_cylinder():
    fig0, ax0 = get_figure(params={'projection': '3d'})
    start, end = np.array([0, 0, 0]), np.array([1, 0, 0])
    plot_cylinder(ax0, start=start, end=end,
                  start_radius=0, end_radius=10.,
                  color='black', alpha=1.)
    assert ax0.has_data()


def test_plot_sphere():
    fig0, ax0 = get_figure(params={'projection': '3d'})
    plot_sphere(ax0, [0, 0, 0], 10., color='black', alpha=1.)
    assert ax0.has_data()



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Morphometrics functions for neurons or neuron populations."""

from functools import partial
import math

import numpy as np

from neurom import morphmath
from neurom.core.neuron import iter_neurites, iter_segments
from neurom.core.dataformat import COLS
from neurom.core.types import NeuriteType
from neurom.core.types import tree_type_checker as is_type
from neurom.features import feature

from neurom.geom import bounding_box

feature = partial(feature, namespace='NEURONFEATURES')


def neuron_population(nrns):
    """Makes sure `nrns` behaves like a neuron population."""
    return nrns.neurons if hasattr(nrns, 'neurons') else (nrns,)


@feature(shape=())
def soma_volume(nrn):
    """Get the volume of a neuron's soma."""
    return nrn.soma.volume


@feature(shape=(...,))
def soma_volumes(nrn_pop):
    """Get the volume of the somata in a population of neurons.

    Note:
        If a single neuron is passed, a single element list with the volume
        of its soma member is returned.
    """
    nrns = neuron_population(nrn_pop)
    return [soma_volume(n) for n in nrns]


@feature(shape=())
def soma_surface_area(nrn, neurite_type=NeuriteType.soma):
    """Get the surface area of a neuron's soma.

    Note:
        The surface area is calculated by assuming the soma is spherical.
    """
    assert neurite_type == NeuriteType.soma, 'Neurite type must be soma'
    return 4 * math.pi * nrn.soma.radius ** 2


@feature(shape=(...,))
def soma_surface_areas(nrn_pop, neurite_type=NeuriteType.soma):
    """Get the surface areas of the somata in a population of neurons.

    Note:
        The surface area is calculated by assuming the soma is spherical.

    Note:
        If a single neuron is passed, a single element list with the surface
        area of its soma member is returned.
    """
    nrns = neuron_population(nrn_pop)
    assert neurite_type == NeuriteType.soma, 'Neurite type must be soma'
    return [soma_surface_area(n) for n in nrns]


@feature(shape=(...,))
def soma_radii(nrn_pop, neurite_type=NeuriteType.soma):
    """Get the radii of the somata of a population of neurons.

    Note:
        If a single neuron is passed, a single element list with the
        radius of its soma member is returned.
    """
    assert neurite_type == NeuriteType.soma, 'Neurite type must be soma'
    nrns = neuron_population(nrn_pop)
    return [n.soma.radius for n in nrns]


@feature(shape=(...,))
def trunk_section_lengths(nrn, neurite_type=NeuriteType.all):
    """List of lengths of trunk sections of neurites in a neuron."""
    neurite_filter = is_type(neurite_type)
    return [morphmath.section_length(s.root_node.points)
            for s in nrn.neurites if neurite_filter(s)]


@feature(shape=(...,))
def trunk_origin_radii(nrn, neurite_type=NeuriteType.all):
    """Radii of the trunk sections of neurites in a neuron."""
    neurite_filter = is_type(neurite_type)
    return [s.root_node.points[0][COLS.R] for s in nrn.neurites if neurite_filter(s)]


@feature(shape=(...,))
def trunk_origin_azimuths(nrn, neurite_type=NeuriteType.all):
    """Get a list of all the trunk origin azimuths of a neuron or population.

    The azimuth is defined as Angle between x-axis and the vector
    defined by (initial tree point - soma center) on the x-z plane.

    The range of the azimuth angle [-pi, pi] radians
    """
    neurite_filter = is_type(neurite_type)
    nrns = neuron_population(nrn)

    def _azimuth(section, soma):
        """Azimuth of a section."""
        vector = morphmath.vector(section[0], soma.center)
        return np.arctan2(vector[COLS.Z], vector[COLS.X])

    return [_azimuth(s.root_node.points, n.soma)
            for n in nrns
            for s in n.neurites if neurite_filter(s)]


@feature(shape=(...,))
def trunk_origin_elevations(nrn, neurite_type=NeuriteType.all):
    """Get a list of all the trunk origin elevations of a neuron or population.

    The elevation is defined as the angle between x-axis and the
    vector defined by (initial tree point - soma center)
    on the x-y half-plane.

    The range of the elevation angle [-pi/2, pi/2] radians
    """
    neurite_filter = is_type(neurite_type)
    nrns = neuron_population(nrn)

    def _elevation(section, soma):
        """Elevation of a section."""
        vector = morphmath.vector(section[0], soma.center)
        norm_vector = np.linalg.norm(vector)

        if norm_vector >= np.finfo(type(norm_vector)).eps:
            return np.arcsin(vector[COLS.Y] / norm_vector)
        raise ValueError("Norm of vector between soma center and section is almost zero.")

    return [_elevation(s.root_node.points, n.soma)
            for n in nrns
            for s in n.neurites if neurite_filter(s)]


@feature(shape=(...,))
def trunk_vectors(nrn, neurite_type=NeuriteType.all):
    """Calculates the vectors between all the trunks of the neuron and the soma center."""
    neurite_filter = is_type(neurite_type)
    nrns = neuron_population(nrn)

    return np.array([morphmath.vector(s.root_node.points[0], n.soma.center)
                     for n in nrns
                     for s in n.neurites if neurite_filter(s)])


@feature(shape=(...,))
def trunk_angles(nrn, neurite_type=NeuriteType.all):
    """Calculates the angles between all the trunks of the neuron.

    The angles are defined on the x-y plane and the trees
    are sorted from the y axis and anticlock-wise.
    """
    vectors = trunk_vectors(nrn, neurite_type=neurite_type)
    # In order to avoid the failure of the process in case the neurite_type does not exist
    if not vectors.size:
        return []

    def _sort_angle(p1, p2):
        """Angle between p1-p2 to sort vectors."""
        ang1 = np.arctan2(*p1[::-1])
        ang2 = np.arctan2(*p2[::-1])
        return ang1 - ang2

    # Sorting angles according to x-y plane
    order = np.argsort(np.array([_sort_angle(i / np.linalg.norm(i), [0, 1])
                                 for i in vectors[:, 0:2]]))

    ordered_vectors = vectors[order][:, [COLS.X, COLS.Y]]

    return [morphmath.angle_between_vectors(ordered_vectors[i], ordered_vectors[i - 1])
            for i, _ in enumerate(ordered_vectors)]


def sholl_crossings(neurites, center, radii):
    """Calculate crossings of neurites.

    This function can also be used with a list aa neurites, as follow:

        secs = (sec for sec in nm.iter_sections(neuron) if complex_filter(sec))
        sholl = nm.features.neuronfunc.sholl_crossings(secs,
                                                       center=neuron.soma.center,
                                                       radii=np.arange(0, 1000, 100))

    Args:
        neurites(list): morphology on which to perform Sholl analysis, or list of neurites
        center(Point): center point
        radii(iterable of floats): radii for which crossings will be counted

    Returns:
        Array of same length as radii, with a count of the number of crossings
        for the respective radius
    """
    def _count_crossings(neurite, radius):
        """Used to count_crossings of segments in neurite with radius."""
        r2 = radius ** 2
        count = 0
        for start, end in iter_segments(neurite):
            start_dist2, end_dist2 = (morphmath.point_dist2(center, start),
                                      morphmath.point_dist2(center, end))

            count += int(start_dist2 <= r2 <= end_dist2 or
                         end_dist2 <= r2 <= start_dist2)

        return count

    return np.array([sum(_count_crossings(neurite, r)
                         for neurite in iter_neurites(neurites))
                     for r in radii])


@feature(shape=(...,))
def sholl_frequency(nrn, neurite_type=NeuriteType.all, step_size=10):
    """Perform Sholl frequency calculations on a population of neurites.

    Args:
        nrn(morph): nrn or population
        neurite_type(NeuriteType): which neurites to operate on
        step_size(float): step size between Sholl radii

    Note:
        Given a neuron, the soma center is used for the concentric circles,
        which range from the soma radii, and the maximum radial distance
        in steps of `step_size`.  When a population is given, the concentric
        circles range from the smallest soma radius to the largest radial neurite
        distance.  Finally, each segment of the neuron is tested, so a neurite that
        bends back on itself, and crosses the same Sholl radius will get counted as
        having crossed multiple times.
    """
    nrns = neuron_population(nrn)
    neurite_filter = is_type(neurite_type)

    min_soma_edge = float('Inf')
    max_radii = 0
    neurites_list = []
    for neuron in nrns:
        neurites_list.extend(((neurites, neuron.soma.center)
                              for neurites in neuron.neurites
                              if neurite_filter(neurites)))

        min_soma_edge = min(min_soma_edge, neuron.soma.radius)
        max_radii = max(max_radii, np.max(np.abs(bounding_box(neuron))))

    radii = np.arange(min_soma_edge, max_radii + step_size, step_size)
    ret = np.zeros_like(radii)
    for neurites, center in neurites_list:
        ret += sholl_crossings(neurites, center, radii)

    return ret



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from pathlib import Path

from neurom.apps import get_config
from neurom.exceptions import ConfigError

import pytest


def test_get_config():
    # get the default

    test_yaml = Path(__file__).parent.parent.parent / 'neurom/apps/config/morph_stats.yaml'

    expected = {'neurite': {'section_lengths': ['max', 'total'], 'section_volumes': ['total'], 'section_branch_orders': ['max']}, 'neurite_type': ['AXON', 'APICAL_DENDRITE', 'BASAL_DENDRITE', 'ALL'], 'neuron': {'soma_radii': ['mean']}}

    config = get_config(None, test_yaml)
    assert config == expected

    config = get_config(test_yaml, '')
    assert config == expected


def test_get_config_exception():
    """current python file isn't a yaml file."""
    with pytest.raises(ConfigError):
        get_config(__file__, {})



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from copy import deepcopy
from io import StringIO
from pathlib import Path

import numpy as np
from neurom import load_neuron
from neurom.check import morphtree as mt
from neurom.core.dataformat import COLS

DATA_PATH = Path(__file__).parent.parent / 'data'
SWC_PATH = DATA_PATH / 'swc'


def _make_flat(neuron):

    class Flattenizer:
        def __call__(self, points):
            points = deepcopy(points)
            points[:, COLS.Z] = 0.
            return points

    return neuron.transform(Flattenizer())


def _make_monotonic(neuron):
    for neurite in neuron.neurites:
        for section in neurite.iter_sections():
            points = section.points
            if section.parent is not None:
                points[0][COLS.R] = section.parent.points[-1][COLS.R] / 2
            for point_id in range(len(points) - 1):
                points[point_id + 1][COLS.R] = points[point_id][COLS.R] / 2.
            section.points = points


def _generate_back_track_tree(n, dev):
    points = np.array(dev) + np.array([1, 3 if n == 0 else -3, 0])

    neuron = load_neuron(StringIO(u"""
    ((CellBody)
     (0 0 0 0.4))

    ((Dendrite)
    (0 0 0 0.4)
    (0 1 0 0.3)
    (0 2 0 0.28)
    (
      (0 2 0 0.28)
      (1 3 0 0.3)
      (2 4 0 0.22)
      |
      (0 2 0 0.28)
      (1 -3 0 0.3)
      (2 -4 0 0.24)
      ({0} {1} {2} 0.52)
      (3 -5 0 0.2)
      (4 -6 0 0.2)
    ))
    """.format(*points.tolist())), reader='asc')

    return neuron


def test_is_monotonic():
    # tree with decreasing radii
    neuron = load_neuron(StringIO(u"""
        ((Dendrite)
        (0 0 0 1.0)
        (0 0 0 0.99)
        (
          (0 0 0 0.99)
          (0 0 0 0.1)
          |
          (0 0 0 0.5)
          (0 0 0 0.2)
        ))"""), reader='asc')
    assert mt.is_monotonic(neuron.neurites[0], 1e-6)

    # tree with equal radii
    neuron = load_neuron(StringIO(u"""
        ((Dendrite)
        (0 0 0 1.0)
        (0 0 0 1.0)
        (
          (0 0 0 1.0)
          (0 0 0 1.0)
          |
          (0 0 0 1.0)
          (0 0 0 1.0)
        ))"""), reader='asc')
    assert mt.is_monotonic(neuron.neurites[0], 1e-6)

    # tree with increasing radii
    neuron = load_neuron(StringIO(u"""
        ((Dendrite)
        (0 0 0 1.0)
        (0 0 0 1.0)
        (
          (0 0 0 1.1)
          (0 0 0 1.1)
          |
          (0 0 0 0.3)
          (0 0 0 0.1)
        ))"""), reader='asc')
    assert not mt.is_monotonic(neuron.neurites[0], 1e-6)

    # Tree with larger child initial point
    neuron = load_neuron(StringIO(u"""
        ((Dendrite)
        (0 0 0 1.0)
        (0 0 0 0.75)
        (0 0 0 0.5)
        (0 0 0 0.25)
        (
          (0 0 0 0.375)
          (0 0 0 0.125)
          (0 0 0 0.625)
        ))"""), reader='asc')
    assert not mt.is_monotonic(neuron.neurites[0], 1e-6)


def test_is_flat():
    neu_tree = load_neuron(Path(SWC_PATH, 'Neuron.swc'))
    assert not mt.is_flat(neu_tree.neurites[0], 1e-6, method='tolerance')
    assert not mt.is_flat(neu_tree.neurites[0], 0.1, method='ratio')


def test_is_back_tracking():
    # case 1: a back-track falls directly on a previous node
    t = _generate_back_track_tree(1, (0., 0., 0.))
    assert mt.is_back_tracking(t.neurites[0])

    # case 2: a zigzag is close to another segment
    t = _generate_back_track_tree(1, (0.1, -0.1, 0.02))
    assert mt.is_back_tracking(t.neurites[0])

    # case 3: a zigzag is close to another segment 2
    t = _generate_back_track_tree(1, (-0.2, 0.04, 0.144))
    assert mt.is_back_tracking(t.neurites[0])

    # case 4: a zigzag far from civilization
    t = _generate_back_track_tree(1, (10., -10., 10.))
    assert not mt.is_back_tracking(t.neurites[0])

    # case 5: a zigzag on another section
    # currently zigzag is defined on the same section
    # thus this test should not be true
    t = _generate_back_track_tree(0, (-0.2, 0.04, 0.144))
    assert not mt.is_back_tracking(t.neurites[0])


def test_get_flat_neurites():
    n = load_neuron(Path(SWC_PATH, 'Neuron.swc'))
    assert len(mt.get_flat_neurites(n, 1e-6, method='tolerance')) == 0
    assert len(mt.get_flat_neurites(n, 0.1, method='ratio')) == 0

    n = _make_flat(n)
    assert len(mt.get_flat_neurites(n, 1e-6, method='tolerance')) == 4
    assert len(mt.get_flat_neurites(n, 0.1, method='ratio')) == 4


def test_get_nonmonotonic_neurites():
    n = load_neuron(Path(SWC_PATH, 'Neuron.swc'))
    assert len(mt.get_nonmonotonic_neurites(n)) == 4
    _make_monotonic(n)
    assert len(mt.get_nonmonotonic_neurites(n)) == 0


def test_get_back_tracking_neurites():
    n = load_neuron(Path(SWC_PATH, 'Neuron.swc'))
    assert len(mt.get_back_tracking_neurites(n)) == 4



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""The morph-tool command line launcher."""
import logging

import click
import matplotlib.pyplot as plt

import neurom as nm
from neurom.apps import morph_stats, morph_check
from neurom import load_neuron
from neurom.viewer import draw as pyplot_draw


@click.group()
@click.option('-v', '--verbose', count=True, default=0,
              help='-v for WARNING, -vv for INFO, -vvv for DEBUG')
def cli(verbose):
    """The CLI entry point."""
    level = (logging.WARNING, logging.INFO, logging.DEBUG)[min(verbose, 2)]
    logging.basicConfig(level=level)


@cli.command()
@click.argument('input_file')
@click.option('--plane', type=click.Choice(['3d', 'xy', 'yx', 'yz', 'zy', 'xz', 'zx']),
              default='3d')
@click.option('--backend', type=click.Choice(['plotly', 'matplotlib']),
              default='matplotlib')
@click.option('-r', '--realistic-diameters/--no-realistic-diameters', default=False,
              help='Scale diameters according to the plot axis\n'
                   'Warning: Only works with the matplotlib backend')
def view(input_file, plane, backend, realistic_diameters):
    """A simple neuron viewer."""
    # pylint: disable=import-outside-toplevel
    if backend == 'matplotlib':
        kwargs = {
            'mode': '3d' if plane == '3d' else '2d',
            'realistic_diameters': realistic_diameters,
        }
        if plane != '3d':
            kwargs['plane'] = plane
        pyplot_draw(load_neuron(input_file), **kwargs)
        plt.show()
    else:
        from neurom.view.plotly import draw as plotly_draw
        plotly_draw(load_neuron(input_file), plane=plane)


@cli.command(short_help='Morphology statistics extractor, more details at'
                        'https://neurom.readthedocs.io/en/latest/morph_stats.html')
@click.argument('datapath', required=False)
@click.option('-C', '--config', type=click.Path(exists=True, dir_okay=False),
              default=morph_stats.EXAMPLE_CONFIG, show_default=True,
              help='Configuration File')
@click.option('-o', '--output', type=click.Path(exists=False, dir_okay=False),
              help='Path to output file, if it ends in .json, a json file is created,'
                   'otherwise a csv file is created')
@click.option('-f', '--full-config', is_flag=True, default=False,
              help='If passed then --config is ignored. Compute statistics for all neurite'
                   'types, all modes and all features')
@click.option('--as-population', is_flag=True, default=False,
              help='If enabled the directory is treated as a population')
@click.option('-I', '--ignored-exceptions', help='Exception to ignore',
              type=click.Choice(morph_stats.IGNORABLE_EXCEPTIONS.keys()))
def stats(datapath, config, output, full_config, as_population, ignored_exceptions):
    """Cli for apps/morph_stats."""
    morph_stats.main(datapath, config, output, full_config, as_population, ignored_exceptions)


@cli.command(short_help='list all available features')
def features():
    """Cli to get list of available features. For backward compatibility."""
    # TODO replace it with programmatically generated Sphinx page that contains all available
    # features, also programmatically generate EXAMPLE_CONFIG on that page.
    # pylint: disable=protected-access
    print(nm.features._get_doc())


@cli.command(short_help='Perform checks on morphologies, more details at'
                        'https://neurom.readthedocs.io/en/latest/morph_check.html')
@click.argument('datapath')
@click.option('-C', '--config', type=click.Path(exists=True, dir_okay=False),
              default=morph_check.EXAMPLE_CONFIG, show_default=True,
              help='Configuration File')
@click.option('-o', '--output', type=click.Path(exists=False, dir_okay=False),
              help='Path to output json summary file')
def check(datapath, config, output):
    """Cli for apps/morph_check."""
    morph_check.main(datapath, config, output)



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test neurom.utils."""
import json
import random
import warnings

import numpy as np
from neurom import utils as nu
import pytest


def test_memoize_caches():
    class A:
        @nu.memoize
        def dummy(self, x, y=42):
            return random.random()

    a = A()
    ref1 = a.dummy(42)
    ref2 = a.dummy(42, 43)
    ref3 = a.dummy(42, y=43)

    for _ in range(10):
        assert a.dummy(42) == ref1
        assert A().dummy(42) != ref1
        assert a.dummy(42, 43) == ref2
        assert A().dummy(42, 43) != ref2
        assert a.dummy(42, y=43) == ref3
        assert A().dummy(42, y=43) != ref3


def test_deprecated():
    @nu.deprecated(msg='Hello')
    def dummy():
        pass

    with warnings.catch_warnings(record=True) as s:
        dummy()
        assert len(s) > 0
        assert s[0].message.args[0] == 'Call to deprecated function dummy. Hello'


def test_deprecated_module():
    with warnings.catch_warnings(record=True) as s:
        nu.deprecated_module('foo', msg='msg')
        assert len(s) > 0


def test_NeuromJSON():
    ex = {'zero': 0,
          'one': np.int64(1),
          'two': np.float32(2.0),
          'three': np.array([1, 2, 3])
          }
    output = json.dumps(ex, cls=nu.NeuromJSON)
    loaded = json.loads(output)
    assert (loaded ==
           {'zero': 0,
            'one': 1,
            'two': 2.0,
            'three': [1, 2, 3]
            })

    enc = nu.NeuromJSON()
    assert enc.default(ex['one']) == 1
    assert enc.default(ex['two']) == 2.0

    with pytest.raises(TypeError):
        enc.default(0)


def test_ordered_enum():
    class Grade(nu.OrderedEnum):
        A = 5
        B = 4
        C = 3
        D = 2
        F = 1

    assert Grade.C < Grade.A
    assert Grade.C <= Grade.A
    assert Grade.C <= Grade.C
    assert not Grade.C > Grade.A
    assert not Grade.C >= Grade.A
    assert Grade.C >= Grade.C
    with pytest.raises(NotImplementedError):
        Grade.__ge__(Grade.A, 1)
    with pytest.raises(NotImplementedError):
        Grade.__le__(Grade.A, 1)
    with pytest.raises(NotImplementedError):
        Grade.__gt__(Grade.A, 1)
    with pytest.raises(NotImplementedError):
        Grade.__lt__(Grade.A, 1)



from pathlib import Path
from neurom import load_neuron
from neurom.apps.annotate import annotate, generate_annotation
from neurom.check import CheckResult
from neurom.check.neuron_checks import has_no_narrow_start

SWC_PATH = Path(__file__).parent.parent / 'data/swc'


def test_generate_annotation():
    checker_ok = CheckResult(True)
    checker_not_ok = CheckResult(False, [('section 1', [[1, 2, 3], [4, 5, 6]]),
                                         ('section 2', [[7, 8, 9], [10, 11, 12]])])

    settings = {'color': 'blue', 'label': 'circle', 'name': 'dangling'}
    assert generate_annotation(checker_ok, settings) == ""

    correct_result = """

(circle   ; MUK_ANNOTATION
    (Color blue)   ; MUK_ANNOTATION
    (Name "dangling")   ; MUK_ANNOTATION
    (      1.00       2.00       3.00 0.50)   ; MUK_ANNOTATION
    (      4.00       5.00       6.00 0.50)   ; MUK_ANNOTATION
    (      7.00       8.00       9.00 0.50)   ; MUK_ANNOTATION
    (     10.00      11.00      12.00 0.50)   ; MUK_ANNOTATION
)   ; MUK_ANNOTATION
"""

    assert generate_annotation(checker_not_ok, settings) == correct_result


def test_annotate():

    correct_result = """

(Circle1   ; MUK_ANNOTATION
    (Color Blue)   ; MUK_ANNOTATION
    (Name "narrow start")   ; MUK_ANNOTATION
    (      0.00       0.00       2.00 0.50)   ; MUK_ANNOTATION
)   ; MUK_ANNOTATION
"""

    checkers = {has_no_narrow_start: {"name": "narrow start",
                                      "label": "Circle1",
                                      "color": "Blue"}}

    neuron = load_neuron(SWC_PATH / 'narrow_start.swc')
    results = [checker(neuron) for checker in checkers.keys()]
    assert annotate(results, checkers.values()) == correct_result



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
from pathlib import Path

import neurom as nm
from neurom.core.neuron import Neurite

from numpy.testing import assert_almost_equal

SWC_PATH = Path(__file__).parent.parent / 'data/swc/'
nrn = nm.load_neuron(SWC_PATH / 'point_soma_single_neurite.swc')

ROOT_NODE = nrn.neurites[0].morphio_root_node
RADIUS = .5
REF_LEN = 3


def test_init():
    nrt = Neurite(ROOT_NODE)
    assert nrt.type == nm.NeuriteType.basal_dendrite
    assert len(nrt.points) == 4


def test_neurite_length():
    nrt = Neurite(ROOT_NODE)
    assert_almost_equal(nrt.length, REF_LEN)


def test_neurite_area():
    nrt = Neurite(ROOT_NODE)
    area = 2 * math.pi * RADIUS * REF_LEN
    assert_almost_equal(nrt.area, area)


def test_neurite_volume():
    nrt = Neurite(ROOT_NODE)
    volume = math.pi * RADIUS * RADIUS * REF_LEN
    assert_almost_equal(nrt.volume, volume)


def test_str():
    nrt = Neurite(ROOT_NODE)
    assert 'Neurite' in str(nrt)


def test_neurite_hash():
    nrt = Neurite(ROOT_NODE)
    assert hash(nrt) == hash((nrt.type, nrt.root_node))



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import warnings
from io import StringIO
from pathlib import Path

import numpy as np
from neurom import load_neuron
from neurom.core.types import NeuriteType
from neurom.view import common, view

import pytest
from numpy.testing import assert_allclose, assert_array_almost_equal

DATA_PATH = Path(__file__).parent.parent / 'data'
SWC_PATH = DATA_PATH / 'swc'
fst_neuron = load_neuron(SWC_PATH / 'Neuron.swc')
simple_neuron = load_neuron(SWC_PATH / 'simple.swc')
neuron_different = load_neuron(SWC_PATH / 'simple-different-section-types.swc')
tree_colors = {'black': np.array([[0., 0., 0., 1.] for _ in range(3)]),
               None: [[1., 0., 0., 1.],
                      [1., 0., 0., 1.],
                      [0.501961, 0., 0.501961, 1.]]}


def test_tree_diameter_scale(get_fig_2d):
    fig, ax = get_fig_2d
    tree = neuron_different.neurites[0]
    for input_color, expected_colors in tree_colors.items():
        view.plot_tree(ax, tree, color=input_color, diameter_scale=None, alpha=1., linewidth=1.2)
        collection = ax.collections[0]
        assert collection.get_linewidth()[0] == 1.2
        assert_array_almost_equal(collection.get_colors(), expected_colors)
        fig.clear()


def test_tree_diameter_real(get_fig_2d):
    fig, ax = get_fig_2d
    tree = neuron_different.neurites[0]
    for input_color, expected_colors in tree_colors.items():
        view.plot_tree(ax, tree, color=input_color, alpha=1., linewidth=1.2, realistic_diameters=True)
        collection = ax.collections[0]
        assert collection.get_linewidth()[0] == 1.0
        assert_array_almost_equal(collection.get_facecolors(), expected_colors)
        fig.clear()


def test_tree_invalid(get_fig_2d):
    fig, ax = get_fig_2d
    with pytest.raises(AssertionError):
        view.plot_tree(ax, neuron_different.neurites[0], plane='wrong')


def test_tree_bounds(get_fig_2d):
    fig, ax = get_fig_2d
    view.plot_tree(ax, neuron_different.neurites[0])
    np.testing.assert_allclose(ax.dataLim.bounds, (-5., 0., 11., 5.))


def test_neuron(get_fig_2d):
    fig, ax = get_fig_2d
    view.plot_neuron(ax, fst_neuron)
    assert ax.get_title() == fst_neuron.name
    assert_allclose(ax.dataLim.get_points(),
                               [[-40.32853516, -57.600172],
                                [64.74726272, 48.51626225], ])

    with pytest.raises(AssertionError):
        view.plot_tree(ax, fst_neuron, plane='wrong')


def test_tree3d(get_fig_3d):
    fig, ax = get_fig_3d
    tree = simple_neuron.neurites[0]
    view.plot_tree3d(ax, tree)
    xy_bounds = ax.xy_dataLim.bounds
    np.testing.assert_allclose(xy_bounds, (-5., 0., 11., 5.))
    zz_bounds = ax.zz_dataLim.bounds
    np.testing.assert_allclose(zz_bounds, (0., 0., 1., 1.))


def test_neuron3d(get_fig_3d):
    fig, ax = get_fig_3d
    view.plot_neuron3d(ax, fst_neuron)
    assert ax.get_title() == fst_neuron.name
    assert_allclose(ax.xy_dataLim.get_points(),
                               [[-40.32853516, -57.600172],
                                [64.74726272, 48.51626225], ])
    assert_allclose(ax.zz_dataLim.get_points().T[0],
                               (-00.09999862, 54.20408797))


def test_neuron_no_neurites(get_fig_2d):
    filename = Path(SWC_PATH, 'point_soma.swc')
    fig, ax = get_fig_2d
    view.plot_neuron(ax, load_neuron(filename))


def test_neuron3d_no_neurites(get_fig_3d):
    filename = Path(SWC_PATH, 'point_soma.swc')
    fig, ax = get_fig_3d
    view.plot_neuron3d(ax, load_neuron(filename))


def test_dendrogram(get_fig_2d):
    fig, ax = get_fig_2d
    view.plot_dendrogram(ax, fst_neuron)
    assert_allclose(ax.get_xlim(), (-10., 180.), rtol=0.25)

    view.plot_dendrogram(ax, fst_neuron, show_diameters=False)
    assert_allclose(ax.get_xlim(), (-10., 180.), rtol=0.25)


with warnings.catch_warnings(record=True):
    soma0 = fst_neuron.soma

    # upright, varying radius
    soma_2pt_normal = load_neuron(StringIO(u"""1 1 0  0 0 1  -1
                                               2 1 0 10 0 10  1"""), reader='swc').soma

    # upright, uniform radius, multiple cylinders
    soma_3pt_normal = load_neuron(StringIO(u"""1 1 0 -10 0 10  -1
                                               2 1 0   0 0 10   1
                                               3 1 0  10 0 10   2"""), reader='swc').soma

    # increasing radius, multiple cylinders
    soma_4pt_normal_cylinder = load_neuron(StringIO(u"""1 1   0   0   0 1 -1
                                                       2 1   0 -10   0 2  1
                                                       3 1   0 -10  10 4  2
                                                       4 1 -10 -10 -10 4  3"""), reader='swc').soma

    soma_4pt_normal_contour = load_neuron(StringIO(u"""((CellBody)
                                                       (0     0   0 1)
                                                       (0   -10   0 2)
                                                       (0   -10  10 4)
                                                       (-10 -10 -10 4))"""), reader='asc').soma


def test_soma(get_fig_2d):
    fig, ax = get_fig_2d
    for s in (soma0,
              soma_3pt_normal,
              soma_4pt_normal_cylinder,
              soma_4pt_normal_contour):
        view.plot_soma(ax, s)
        common.plt.close(fig)

        view.plot_soma(ax, s, soma_outline=False)
        common.plt.close(fig)


def test_soma3d(get_fig_3d):
    _, ax = get_fig_3d
    view.plot_soma3d(ax, soma_3pt_normal)
    assert_allclose(ax.get_xlim(), (-11.,  11.), atol=2)
    assert_allclose(ax.get_ylim(), (-11.,  11.), atol=2)
    assert_allclose(ax.get_zlim(), (-10.,  10.), atol=2)


def test_get_color():
    assert view._get_color(None, NeuriteType.basal_dendrite) == "red"
    assert view._get_color(None, NeuriteType.axon) == "blue"
    assert view._get_color(None, NeuriteType.apical_dendrite) == "purple"
    assert view._get_color(None, NeuriteType.soma) == "black"
    assert view._get_color(None, NeuriteType.undefined) == "green"
    assert view._get_color(None, 'wrong') == "green"
    assert view._get_color('blue', 'wrong') == "blue"
    assert view._get_color('yellow', NeuriteType.axon) == "yellow"



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""compare neurom.features features with values dumped from the original
neurom._point_neurite.features"""

import json
import warnings
from itertools import chain
from pathlib import Path

import neurom as nm
from neurom.core.neuron import Section
from neurom.core.types import NeuriteType
from neurom.features import bifurcationfunc as _bf
from neurom.features import neuritefunc as _nrt
from neurom.features import neuronfunc as _nrn
from neurom.features import sectionfunc as _sec

from numpy.testing import assert_almost_equal
from utils import _close, _equal

DATA_PATH =  Path(__file__).parent.parent / 'data'
SWC_DATA_PATH = DATA_PATH / 'swc'
H5V1_DATA_PATH = DATA_PATH / 'h5/v1'
H5V2_DATA_PATH = DATA_PATH / 'h5/v2'
MORPH_FILENAME = 'Neuron.h5'
SWC_MORPH_FILENAME = 'Neuron.swc'

REF_NEURITE_TYPES = [NeuriteType.apical_dendrite, NeuriteType.basal_dendrite,
                     NeuriteType.basal_dendrite, NeuriteType.axon]

json_data = json.load(open(DATA_PATH / 'dataset/point_neuron_feature_values.json'))


def get(feat, neurite_format, **kwargs):
    """using the values captured from the old point_neurite system."""
    neurite_type = str(kwargs.get('neurite_type', ''))
    return json_data[neurite_format][feat][neurite_type]


def i_chain2(trees, iterator_type=Section.ipreorder, mapping=None, tree_filter=None):
    """Returns a mapped iterator to a collection of trees
    Provides access to all the elements of all the trees
    in one iteration sequence.
    Parameters:
        trees: iterator or iterable of tree objects
        iterator_type: type of the iteration (segment, section, triplet...)
        mapping: optional function to apply to the iterator's target.
        tree_filter: optional top level filter on properties of tree objects.
    """
    nrt = (trees if tree_filter is None
           else filter(tree_filter, trees))

    chain_it = chain.from_iterable(map(iterator_type, nrt))
    return chain_it if mapping is None else map(mapping, chain_it)


class SectionTreeBase:
    """Base class for section tree tests."""

    def setup_method(self):
        self.ref_nrn = 'h5'
        self.ref_types = REF_NEURITE_TYPES

    def test_neurite_type(self):
        neurite_types = [n0.type for n0 in self.sec_nrn.neurites]
        assert neurite_types == self.ref_types

    def test_get_n_sections(self):
        assert (_nrt.n_sections(self.sec_nrn) ==
                        get('number_of_sections', self.ref_nrn)[0])

        for t in NeuriteType:
            actual = _nrt.n_sections(self.sec_nrn, neurite_type=t)
            assert (actual ==
                            get('number_of_sections', self.ref_nrn, neurite_type=t)[0])

    def test_get_number_of_sections_per_neurite(self):
        _equal(_nrt.number_of_sections_per_neurite(self.sec_nrn),
               get('number_of_sections_per_neurite', self.ref_nrn))

        for t in NeuriteType:
            _equal(_nrt.number_of_sections_per_neurite(self.sec_nrn, neurite_type=t),
                   get('number_of_sections_per_neurite', self.ref_nrn, neurite_type=t))

    def test_get_n_segments(self):
        assert _nrt.n_segments(self.sec_nrn) == get('number_of_segments', self.ref_nrn)[0]
        for t in NeuriteType:
            assert (_nrt.n_segments(self.sec_nrn, neurite_type=t) ==
                            get('number_of_segments', self.ref_nrn, neurite_type=t)[0])

    def test_get_number_of_neurites(self):
        assert _nrt.n_neurites(self.sec_nrn) == get('number_of_neurites', self.ref_nrn)[0]
        for t in NeuriteType:
            assert (_nrt.n_neurites(self.sec_nrn, neurite_type=t) ==
                            get('number_of_neurites', self.ref_nrn, neurite_type=t)[0])

    def test_get_section_path_distances(self):
        _close(_nrt.section_path_lengths(self.sec_nrn), get('section_path_distances', self.ref_nrn))
        for t in NeuriteType:
            _close(_nrt.section_path_lengths(self.sec_nrn, neurite_type=t),
                   get('section_path_distances', self.ref_nrn, neurite_type=t))

        pl = [_sec.section_path_length(s) for s in i_chain2(self.sec_nrn_trees)]
        _close(pl, get('section_path_distances', self.ref_nrn))

    def test_get_soma_radius(self):
        assert_almost_equal(self.sec_nrn.soma.radius, get('soma_radii', self.ref_nrn)[0])

    def test_get_soma_surface_area(self):
        assert_almost_equal(_nrn.soma_surface_area(self.sec_nrn), get('soma_surface_areas', self.ref_nrn)[0])

    def test_get_soma_volume(self):
        assert_almost_equal(_nrn.soma_volume(self.sec_nrn), get('soma_volumes', self.ref_nrn)[0])

    def test_get_local_bifurcation_angles(self):
        _close(_nrt.local_bifurcation_angles(self.sec_nrn),
               get('local_bifurcation_angles', self.ref_nrn))

        for t in NeuriteType:
            _close(_nrt.local_bifurcation_angles(self.sec_nrn, neurite_type=t),
                   get('local_bifurcation_angles', self.ref_nrn, neurite_type=t))

        ba = [_bf.local_bifurcation_angle(b)
              for b in i_chain2(self.sec_nrn_trees, iterator_type=Section.ibifurcation_point)]

        _close(ba, get('local_bifurcation_angles', self.ref_nrn))

    def test_get_remote_bifurcation_angles(self):
        _close(_nrt.remote_bifurcation_angles(self.sec_nrn),
               get('remote_bifurcation_angles', self.ref_nrn))

        for t in NeuriteType:
            _close(_nrt.remote_bifurcation_angles(self.sec_nrn, neurite_type=t),
                   get('remote_bifurcation_angles', self.ref_nrn, neurite_type=t))

        ba = [_bf.remote_bifurcation_angle(b)
              for b in i_chain2(self.sec_nrn_trees, iterator_type=Section.ibifurcation_point)]

        _close(ba, get('remote_bifurcation_angles', self.ref_nrn))

    def test_get_section_radial_distances(self):
        _close(_nrt.section_radial_distances(self.sec_nrn),
               get('section_radial_distances', self.ref_nrn))

        for t in NeuriteType:
            _close(_nrt.section_radial_distances(self.sec_nrn, neurite_type=t),
                   get('section_radial_distances', self.ref_nrn, neurite_type=t))

    def test_get_trunk_origin_radii(self):
        assert_almost_equal(_nrn.trunk_origin_radii(self.sec_nrn),
                            get('trunk_origin_radii', self.ref_nrn))
        for t in NeuriteType:
            assert_almost_equal(_nrn.trunk_origin_radii(self.sec_nrn, neurite_type=t),
                                get('trunk_origin_radii', self.ref_nrn, neurite_type=t))

    def test_get_trunk_section_lengths(self):
        _close(_nrn.trunk_section_lengths(self.sec_nrn), get('trunk_section_lengths', self.ref_nrn))
        for t in NeuriteType:
            _close(_nrn.trunk_section_lengths(self.sec_nrn, neurite_type=t),
                   get('trunk_section_lengths', self.ref_nrn, neurite_type=t))


class TestH5V1(SectionTreeBase):

    def setup_method(self):
        super(TestH5V1, self).setup_method()
        self.sec_nrn = nm.load_neuron(Path(H5V1_DATA_PATH, MORPH_FILENAME))
        self.sec_nrn_trees = [n.root_node for n in self.sec_nrn.neurites]

    # Overriding soma values as the same soma points in SWC and ASC have different
    # meanings. Hence leading to different values
    def test_get_soma_radius(self):
        assert_almost_equal(self.sec_nrn.soma.radius, 0.09249506049313666)

    def test_get_soma_surface_area(self):
        assert_almost_equal(_nrn.soma_surface_area(self.sec_nrn),
                        0.1075095256160432)

    def test_get_soma_volume(self):
        with warnings.catch_warnings(record=True):
            assert_almost_equal(_nrn.soma_volume(self.sec_nrn), 0.0033147000251481135)


class TestSWC(SectionTreeBase):

    def setup_method(self):
        self.ref_nrn = 'swc'
        self.sec_nrn = nm.load_neuron(Path(SWC_DATA_PATH, SWC_MORPH_FILENAME))
        self.sec_nrn_trees = [n.root_node for n in self.sec_nrn.neurites]
        self.ref_types = [NeuriteType.axon,
                          NeuriteType.basal_dendrite,
                          NeuriteType.basal_dendrite,
                          NeuriteType.apical_dendrite,
                          ]



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
from pathlib import Path

import neurom.geom.transform as gtr
import numpy as np
from neurom import COLS, load_neuron
from neurom.features import neuritefunc as _nf

import pytest
from numpy.testing import assert_almost_equal

TEST_UVEC = np.array([0.01856633,  0.37132666,  0.92831665])
TEST_ANGLE = np.pi / 3.
DATA_PATH = Path(__file__).parent.parent / 'data'
H5_NRN_PATH = DATA_PATH / 'h5/v1/Neuron.h5'
SWC_NRN_PATH = DATA_PATH / 'swc/Neuron.swc'


def _Rx(angle):
    sn = np.sin(angle)
    cs = np.cos(angle)
    return np.array([[1., 0., 0.],
                     [0., cs, -sn],
                     [0., sn, cs]])


def _Ry(angle):
    sn = np.sin(angle)
    cs = np.cos(angle)
    return np.array([[cs, 0., sn],
                     [0., 1., 0.],
                     [-sn, 0., cs]])


def _Rz(angle):
    sn = np.sin(angle)
    cs = np.cos(angle)
    return np.array([[cs, -sn, 0.],
                     [sn, cs, 0.],
                     [0., 0., 1.]])


def test_not_implemented_transform_call_raises():
    with pytest.raises(NotImplementedError):
        class Dummy(gtr.Transform3D):
            pass

        d = Dummy()
        d([1, 2, 3])


def test_translate_bad_type_raises():
    with pytest.raises(NotImplementedError):
        gtr.translate("hello", [1, 2, 3])


def test_rotate_bad_type_raises():
    with pytest.raises(NotImplementedError):
        gtr.rotate("hello", [1, 0, 0], math.pi)


def test_translate_point():

    t = gtr.Translation([100, -100, 100])
    point = [1, 2, 3]
    assert t(point).tolist() == [101, -98, 103]


def test_translate_points():

    t = gtr.Translation([100, -100, 100])
    points = np.array([[1, 2, 3], [11, 22, 33], [111, 222, 333]])
    assert np.all(t(points) == np.array([[101, -98, 103],
                                                 [111, -78, 133],
                                                 [211, 122, 433]]))


ROT_90 = np.array([[0, -1, 0],
                   [1, 0, 0],
                   [0, 0, 1]])

ROT_180 = np.array([[-1, 0, 0],
                    [0, -1, 0],
                    [0, 0, 1]])

ROT_270 = np.array([[0, 1, 0],
                    [-1, 0, 0],
                    [0, 0, 1]])


def test_rotate_point():

    rot = gtr.Rotation(ROT_90)
    assert rot([2, 0, 0]).tolist() == [0, 2, 0]
    assert rot([0, 2, 0]).tolist() == [-2, 0, 0]
    assert rot([0, 0, 2]).tolist() == [0, 0, 2]

    rot = gtr.Rotation(ROT_180)
    assert rot([2, 0, 0]).tolist() == [-2, 0, 0]
    assert rot([0, 2, 0]).tolist() == [0, -2, 0]
    assert rot([0, 0, 2]).tolist() == [0, 0, 2]

    rot = gtr.Rotation(ROT_270)
    assert rot([2, 0, 0]).tolist() == [0, -2, 0]
    assert rot([0, 2, 0]).tolist() == [2, 0, 0]
    assert rot([0, 0, 2]).tolist() == [0, 0, 2]


def test_rotate_points():

    rot = gtr.Rotation(ROT_90)

    points = np.array([[2, 0, 0],
                       [0, 2, 0],
                       [0, 0, 2],
                       [3, 0, 3]])

    assert np.all(rot(points) == np.array([[0, 2, 0],
                                                   [-2, 0, 0],
                                                   [0, 0, 2],
                                                   [0, 3, 3]]))

    rot = gtr.Rotation(ROT_180)
    assert np.all(rot(points) == np.array([[-2, 0, 0],
                                                   [0, -2, 0],
                                                   [0, 0, 2],
                                                   [-3, 0, 3]]))

    rot = gtr.Rotation(ROT_270)
    assert np.all(rot(points) == np.array([[0, -2, 0],
                                                   [2, 0, 0],
                                                   [0, 0, 2],
                                                   [0, -3, 3]]))


def test_pivot_rotate_point():

    point = [1, 2, 3]

    new_orig = np.array([10., 45., 50.])

    t = gtr.Translation(new_orig)
    t_inv = gtr.Translation(new_orig * -1)

    R = gtr._rodrigues_to_dcm(TEST_UVEC, np.pi)

    # change origin, rotate 180
    p1 = gtr.PivotRotation(R, new_orig)(point)

    # do the steps manually
    p2 = t_inv(point)
    p2 = gtr.Rotation(R)(p2)
    p2 = t(p2)

    assert p1.tolist() == p2.tolist()


def test_pivot_rotate_points():

    points = np.array([[1, 2, 3],
                       [4, 5, 6],
                       [7, 8, 9],
                       [10, 11, 12]])

    new_orig = np.array([10., 45., 50.])

    t = gtr.Translation(new_orig)
    t_inv = gtr.Translation(new_orig * -1)

    R = gtr._rodrigues_to_dcm(TEST_UVEC, np.pi)

    # change origin, rotate 180
    p1 = gtr.PivotRotation(R, new_orig)(points)

    # do the steps manually
    p2 = t_inv(points)
    p2 = gtr.Rotation(R)(p2)
    p2 = t(p2)

    assert np.all(p1 == p2)


def _check_fst_nrn_translate(nrn_a, nrn_b, t):

    # soma points
    assert np.allclose(
        (nrn_b.soma.points[:, COLS.XYZ] - nrn_a.soma.points[:, COLS.XYZ]), t)
    _check_fst_neurite_translate(nrn_a.neurites, nrn_b.neurites, t)


def _check_fst_neurite_translate(nrts_a, nrts_b, t):
    # neurite sections
    for sa, sb in zip(_nf.iter_sections(nrts_a),
                      _nf.iter_sections(nrts_b)):
        assert np.allclose((sb.points[:, COLS.XYZ] - sa.points[:, COLS.XYZ]), t)


def test_translate_fst_neuron_swc():

    t = np.array([100., 100., 100.])
    nrn = load_neuron(SWC_NRN_PATH)
    tnrn = gtr.translate(nrn, t)
    _check_fst_nrn_translate(nrn, tnrn, t)


def test_transform_translate_neuron_swc():
    t = np.array([100., 100., 100.])
    nrn = load_neuron(SWC_NRN_PATH)
    tnrn = nrn.transform(gtr.Translation(t))
    _check_fst_nrn_translate(nrn, tnrn, t)


def test_translate_fst_neuron_h5():

    t = np.array([100., 100., 100.])
    nrn = load_neuron(H5_NRN_PATH)
    tnrn = gtr.translate(nrn, t)

    _check_fst_nrn_translate(nrn, tnrn, t)


def test_transform_translate_neuron_h5():
    t = np.array([100., 100., 100.])
    nrn = load_neuron(H5_NRN_PATH)
    tnrn = nrn.transform(gtr.Translation(t))
    _check_fst_nrn_translate(nrn, tnrn, t)


def _apply_rot(points, rot_mat):
    return np.dot(rot_mat, np.array(points).T).T


def _check_fst_nrn_rotate(nrn_a, nrn_b, rot_mat):

    # soma points
    assert np.allclose(_apply_rot(nrn_a.soma.points[:, COLS.XYZ], rot_mat),
                               nrn_b.soma.points[:, COLS.XYZ])

    # neurite sections
    _check_fst_neurite_rotate(nrn_a.neurites, nrn_b.neurites, rot_mat)


def _check_fst_neurite_rotate(nrt_a, nrt_b, rot_mat):
    for sa, sb in zip(_nf.iter_sections(nrt_a),
                      _nf.iter_sections(nrt_b)):
        assert np.allclose(sb.points[:, COLS.XYZ],
                                   _apply_rot(sa.points[:, COLS.XYZ], rot_mat))


def test_rotate_neuron_swc():
    nrn_a = load_neuron(SWC_NRN_PATH)
    nrn_b = gtr.rotate(nrn_a, [0, 0, 1], math.pi/2.0)
    rot = gtr._rodrigues_to_dcm([0, 0, 1], math.pi/2.0)
    _check_fst_nrn_rotate(nrn_a, nrn_b, rot)


def test_transform_rotate_neuron_swc():
    rot = gtr.Rotation(ROT_90)
    nrn_a = load_neuron(SWC_NRN_PATH)
    nrn_b = nrn_a.transform(rot)
    _check_fst_nrn_rotate(nrn_a, nrn_b, ROT_90)


def test_rotate_neuron_h5():
    nrn_a = load_neuron(H5_NRN_PATH)
    nrn_b = gtr.rotate(nrn_a, [0, 0, 1], math.pi/2.0)
    rot = gtr._rodrigues_to_dcm([0, 0, 1], math.pi/2.0)
    _check_fst_nrn_rotate(nrn_a, nrn_b, rot)


def test_transform_rotate_neuron_h5():
    rot = gtr.Rotation(ROT_90)
    nrn_a = load_neuron(H5_NRN_PATH)
    nrn_b = nrn_a.transform(rot)
    _check_fst_nrn_rotate(nrn_a, nrn_b, ROT_90)


def test_rodrigues_to_dcm():

    RES = np.array([[0.50017235, -0.80049871, 0.33019604],
                    [0.80739289, 0.56894174, 0.15627544],
                    [-0.3129606, 0.18843328, 0.9308859]])

    R = gtr._rodrigues_to_dcm(TEST_UVEC, TEST_ANGLE)

    # assess rotation matrix properties:

    # detR = +=1
    assert_almost_equal(np.linalg.det(R), 1.)

    # R.T = R^-1
    assert np.allclose(np.linalg.inv(R), R.transpose())

    # check against calculated matrix
    assert np.allclose(R, RES)

    # check if opposite sign generates inverse
    Rinv = gtr._rodrigues_to_dcm(TEST_UVEC, -TEST_ANGLE)

    assert np.allclose(np.dot(Rinv, R), np.identity(3))

    # check basic rotations with a range of angles
    for angle in np.linspace(0., 2. * np.pi, 10):

        Rx = gtr._rodrigues_to_dcm(np.array([1., 0., 0.]), angle)
        Ry = gtr._rodrigues_to_dcm(np.array([0., 1., 0.]), angle)
        Rz = gtr._rodrigues_to_dcm(np.array([0., 0., 1.]), angle)

        assert np.allclose(Rx, _Rx(angle))
        assert np.allclose(Ry, _Ry(angle))
        assert np.allclose(Rz, _Rz(angle))



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from copy import copy, deepcopy
from pathlib import Path

import neurom as nm
import numpy as np
from morphio import Morphology as ImmutMorphology
from morphio.mut import Morphology
from neurom.core.neuron import Neuron, graft_neuron, iter_segments
from numpy.testing import assert_array_equal

SWC_PATH = Path(__file__).parent.parent / 'data/swc/'


def test_simple():
    nm.load_neuron(str(SWC_PATH / 'simple.swc'))


def test_load_neuron_pathlib():
    nm.load_neuron(SWC_PATH / 'simple.swc')


def test_load_neuron_from_other_neurons():
    filename = SWC_PATH / 'simple.swc'

    expected_points = [[ 0.,  0.,  0.,  1.],
                       [ 0.,  5.,  0.,  1.],
                       [ 0.,  5.,  0.,  1.],
                       [-5.,  5.,  0.,  0.],
                       [ 0.,  5.,  0.,  1.],
                       [ 6.,  5.,  0.,  0.],
                       [ 0.,  0.,  0.,  1.],
                       [ 0., -4.,  0.,  1.],
                       [ 0., -4.,  0.,  1.],
                       [ 6., -4.,  0.,  0.],
                       [ 0., -4.,  0.,  1.],
                       [-5., -4.,  0.,  0.]]

    assert_array_equal(nm.load_neuron(nm.load_neuron(filename)).points,
                       expected_points)

    assert_array_equal(nm.load_neuron(Morphology(filename)).points,
                       expected_points)

    assert_array_equal(nm.load_neuron(ImmutMorphology(filename)).points,
                       expected_points)


def test_for_morphio():
    Neuron(Morphology())

    neuron = Morphology()
    neuron.soma.points = [[0,0,0], [1,1,1], [2,2,2]]
    neuron.soma.diameters = [1, 1, 1]

    NeuroM_neuron = Neuron(neuron)
    assert_array_equal(NeuroM_neuron.soma.points,
                       [[0., 0., 0., 0.5],
                        [1., 1., 1., 0.5],
                        [2., 2., 2., 0.5]])

    NeuroM_neuron.soma.points = [[1, 1, 1, 1],
                                 [2, 2, 2, 2]]
    assert_array_equal(NeuroM_neuron.soma.points,
                       [[1, 1, 1, 1],
                        [2, 2, 2, 2]])


def _check_cloned_neuron(nrn1, nrn2):
    # check if two neurons are identical

    # soma
    assert isinstance(nrn2.soma, type(nrn1.soma))
    assert nrn1.soma.radius == nrn2.soma.radius

    for v1, v2 in zip(nrn1.soma.iter(), nrn2.soma.iter()):
        assert np.allclose(v1, v2)

    # neurites
    for v1, v2 in zip(iter_segments(nrn1), iter_segments(nrn2)):
        (v1_start, v1_end), (v2_start, v2_end) = v1, v2
        assert np.allclose(v1_start, v2_start)
        assert np.allclose(v1_end, v2_end)

    # check if the ids are different
    # somata
    assert nrn1.soma is not nrn2.soma

    # neurites
    for neu1, neu2 in zip(nrn1.neurites, nrn2.neurites):
        assert neu1 is not neu2

    # check if changes are propagated between neurons
    nrn2.soma.radius = 10.
    assert nrn1.soma.radius != nrn2.soma.radius


def test_copy():
    nrn = nm.load_neuron(SWC_PATH / 'simple.swc')
    _check_cloned_neuron(nrn, copy(nrn))


def test_deepcopy():
    nrn = nm.load_neuron(SWC_PATH / 'simple.swc')
    _check_cloned_neuron(nrn, deepcopy(nrn))


def test_graft_neuron():
    nrn1 = nm.load_neuron(SWC_PATH / 'simple.swc')
    basal_dendrite = nrn1.neurites[0]
    nrn2 = graft_neuron(basal_dendrite.root_node)
    assert len(nrn2.neurites) == 1
    assert basal_dendrite == nrn2.neurites[0]


def test_str():
    n = nm.load_neuron(SWC_PATH / 'simple.swc')
    assert 'Neuron' in str(n)
    assert 'Section' in str(n.neurites[0].root_node)



import json
from pathlib import Path
import tempfile

import pandas as pd
import yaml
from click.testing import CliRunner
from mock import patch

from neurom.apps.cli import cli
from neurom.exceptions import ConfigError

DATA = Path(__file__).parent.parent / 'data'


@patch('neurom.apps.cli.plt.show')
def test_viewer_matplotlib(mock):
    runner = CliRunner()
    filename = str(DATA / 'swc' / 'simple.swc')

    result = runner.invoke(cli, ['view', filename])
    assert result.exit_code == 0
    mock.assert_called_once()

    mock.reset_mock()
    result = runner.invoke(cli, ['view', filename, '--plane', 'xy'])
    assert result.exit_code == 0
    mock.assert_called_once()


@patch('neurom.view.plotly.plot')
def test_viewer_plotly(mock):
    runner = CliRunner()
    filename = str(DATA / 'swc' / 'simple.swc')

    result = runner.invoke(cli, ['view', filename,
                                 '--backend', 'plotly'])
    assert result.exit_code == 0
    mock.assert_called_once()

    mock.reset_mock()
    result = runner.invoke(cli, ['view', filename,
                                 '--backend', 'plotly',
                                 '--plane', 'xy'])
    assert result.exit_code == 0
    mock.assert_called_once()


def test_morph_stat():
    runner = CliRunner()
    filename = DATA / 'swc' / 'simple.swc'
    with tempfile.NamedTemporaryFile() as f:
        result = runner.invoke(cli, ['stats', str(filename), '--output', f.name])
        assert result.exit_code == 0
        df = pd.read_csv(f)
        assert set(df.columns) == {'name', 'axon:max_section_length', 'axon:total_section_length',
                                   'axon:total_section_volume', 'axon:max_section_branch_order',
                                   'apical_dendrite:max_section_length',
                                   'apical_dendrite:total_section_length',
                                   'apical_dendrite:total_section_volume',
                                   'apical_dendrite:max_section_branch_order',
                                   'basal_dendrite:max_section_length',
                                   'basal_dendrite:total_section_length',
                                   'basal_dendrite:total_section_volume',
                                   'basal_dendrite:max_section_branch_order',
                                   'all:max_section_length',
                                   'all:total_section_length', 'all:total_section_volume',
                                   'all:max_section_branch_order', 'neuron:mean_soma_radius'}


def test_morph_stat_full_config():
    runner = CliRunner()
    filename = DATA / 'h5/v1/Neuron.h5'
    with tempfile.NamedTemporaryFile() as f:
        result = runner.invoke(cli, ['stats', str(filename), '--full-config', '--output', f.name])
        assert result.exit_code == 0
        df = pd.read_csv(f)
        assert not df.empty


def test_morph_stat_invalid_config():
    runner = CliRunner()
    with tempfile.NamedTemporaryFile('w') as config_f:
        yaml.dump({'neurite': 'invalid'}, config_f)
        result = runner.invoke(cli, ['stats', '--config', config_f.name])
        assert result.exit_code == 1
        assert isinstance(result.exception, ConfigError)


def test_morph_stat_stdout():
    runner = CliRunner()
    filename = DATA / 'swc' / 'simple.swc'
    result = runner.invoke(cli, ['stats', str(filename)])
    assert result.exit_code == 0


def test_morph_stat_as_population():
    runner = CliRunner()
    filename = DATA / 'swc' / 'simple.swc'
    result = runner.invoke(cli, ['stats', str(filename), '--as-population'])
    assert result.exit_code == 0


def test_morph_stat_json():
    runner = CliRunner()
    filename = DATA / 'swc' / 'simple.swc'
    with tempfile.NamedTemporaryFile(suffix='.json') as f:
        result = runner.invoke(cli, ['stats', str(filename), '--output', f.name])
        assert result.exit_code == 0
        content = json.load(f)
        assert content



def test_morph_check():
    runner = CliRunner()
    filename = DATA / 'swc' / 'simple.swc'
    with tempfile.NamedTemporaryFile() as f:
        result = runner.invoke(cli, ['check', str(filename), '--output', f.name])
        assert result.exit_code == 0
        content = json.load(f)
        assert content == {'files': {
            str(filename.absolute()): {'Has basal dendrite': True,
                                       'Has axon': True,
                                       'Has apical dendrite': False,
                                       'Has all nonzero segment lengths': True,
                                       'Has all nonzero section lengths': True,
                                       'Has all nonzero neurite radii': False,
                                       'Has nonzero soma radius': True,
                                       'ALL': False}},
            'STATUS': 'FAIL'}


def test_features():
    runner = CliRunner()
    result = runner.invoke(cli, ['features'])
    assert result.exit_code == 0
    assert len(result.stdout) > 0



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from copy import deepcopy
from io import StringIO
from pathlib import Path

from neurom import check, load_neuron
from neurom.check import neuron_checks as nrn_chk
from neurom.core.dataformat import COLS
from neurom.core.types import dendrite_filter
from numpy.testing import assert_array_equal

DATA_PATH = Path(__file__).parent.parent / 'data'
SWC_PATH = DATA_PATH / 'swc'
ASC_PATH = DATA_PATH / 'neurolucida'
H5V1_PATH = DATA_PATH / 'h5/v1'



def _load_neuron(name):
    if name.endswith('.swc'):
        path = SWC_PATH / name
    elif name.endswith('.h5'):
        path = H5V1_PATH / name
    else:
        path = ASC_PATH / name
    return name, load_neuron(path)


def _make_monotonic(neuron):
    for neurite in neuron.neurites:
        for node in neurite.iter_sections():
            sec = node.points
            if node.parent is not None:
                sec[0][COLS.R] = node.parent.points[-1][COLS.R] / 2.
            for point_id in range(len(sec) - 1):
                sec[point_id + 1][COLS.R] = sec[point_id][COLS.R] / 2.


def _make_flat(neuron):

    class Flattenizer:
        def __call__(self, points):
            points = deepcopy(points)
            points[:, COLS.Z] = 0.;
            return points

    return neuron.transform(Flattenizer())


NEURONS = dict([_load_neuron(n) for n in ['Neuron.h5',
                                          'Neuron_2_branch.h5',
                                          'Neuron.swc',
                                          'Neuron_small_radius.swc',
                                          'Neuron_zero_length_sections.swc',
                                          'Neuron_zero_length_segments.swc',
                                          'Neuron_zero_radius.swc',
                                          'Single_apical.swc',
                                          'Single_axon.swc',
                                          'Single_basal.swc',
                                          ]])


def _pick(files):
    return [NEURONS[f] for f in files]


def test_has_axon_good_data():
    files = ['Neuron.swc',
             'Neuron_small_radius.swc',
             'Single_axon.swc',
             'Neuron.h5',
             ]
    for n in _pick(files):
        assert nrn_chk.has_axon(n)


def test_has_axon_bad_data():
    files = ['Single_apical.swc',
             'Single_basal.swc',
             ]
    for n in _pick(files):
        assert not nrn_chk.has_axon(n)


def test_has_apical_dendrite_good_data():
    files = ['Neuron.swc',
             'Neuron_small_radius.swc',
             'Single_apical.swc',
             'Neuron.h5',
             ]

    for n in _pick(files):
        assert nrn_chk.has_apical_dendrite(n)


def test_has_apical_dendrite_bad_data():
    files = ['Single_axon.swc',
             'Single_basal.swc',
             ]

    for n in _pick(files):
        assert not nrn_chk.has_apical_dendrite(n)


def test_has_basal_dendrite_good_data():
    files = ['Neuron.swc',
             'Neuron_small_radius.swc',
             'Single_basal.swc',
             'Neuron_2_branch.h5',
             'Neuron.h5',
             ]

    for n in _pick(files):
        assert nrn_chk.has_basal_dendrite(n)


def test_has_basal_dendrite_bad_data():
    files = ['Single_axon.swc',
             'Single_apical.swc',
             ]

    for n in _pick(files):
        assert not nrn_chk.has_basal_dendrite(n)


def test_has_no_flat_neurites():

    _, n = _load_neuron('Neuron.swc')

    assert nrn_chk.has_no_flat_neurites(n, 1e-6, method='tolerance')
    assert nrn_chk.has_no_flat_neurites(n, 0.1, method='ratio')

    n = _make_flat(n)

    assert not nrn_chk.has_no_flat_neurites(n, 1e-6, method='tolerance')
    assert not nrn_chk.has_no_flat_neurites(n, 0.1, method='ratio')


def test_nonzero_neurite_radii_good_data():
    files = ['Neuron.swc',
             'Single_apical.swc',
             'Single_basal.swc',
             'Single_axon.swc',
             'Neuron_2_branch.h5',
             ]

    for n in _pick(files):
        ids = nrn_chk.has_all_nonzero_neurite_radii(n)
        assert len(ids.info) == 0


def test_has_all_nonzero_neurite_radii_threshold():
    nrn = NEURONS['Neuron.swc']

    ids = nrn_chk.has_all_nonzero_neurite_radii(nrn)
    assert ids.status

    ids = nrn_chk.has_all_nonzero_neurite_radii(nrn, threshold=0.25)
    assert len(ids.info) == 122


def test_nonzero_neurite_radii_bad_data():
    nrn = NEURONS['Neuron_zero_radius.swc']
    ids = nrn_chk.has_all_nonzero_neurite_radii(nrn, threshold=0.7)
    assert ids.info == [(0, 2)]


def test_nonzero_segment_lengths_good_data():
    nrn = NEURONS['Neuron.swc']
    ids = nrn_chk.has_all_nonzero_segment_lengths(nrn)
    assert ids.status
    assert len(ids.info) == 0


def test_nonzero_segment_lengths_bad_data():
    files = ['Neuron_zero_length_segments.swc',
             'Single_apical.swc',
             'Single_basal.swc',
             'Single_axon.swc',
             ]

    bad_ids = [[0, 21, 42, 63], [0], [0], [0], [0]]

    for i, nrn in enumerate(_pick(files)):
        ids = nrn_chk.has_all_nonzero_segment_lengths(nrn)
        assert (ids.info ==
                        [(id, 0) for id in bad_ids[i]])


def test_nonzero_segment_lengths_threshold():
    nrn = NEURONS['Neuron.swc']

    ids = nrn_chk.has_all_nonzero_segment_lengths(nrn)
    assert ids.status
    assert len(ids.info) == 0

    ids = nrn_chk.has_all_nonzero_segment_lengths(nrn, threshold=0.25)

    bad_ids = [(0, 0), (21, 0), (36, 9), (42, 0), (52, 7), (60, 2), (63, 0), (70, 4), (76, 6)]
    assert (ids.info ==
                    [(id, val) for id, val in bad_ids])


def test_nonzero_section_lengths_good_data():
    files = ['Neuron.swc',
             'Single_apical.swc',
             'Single_basal.swc',
             'Single_axon.swc',
             ]

    for i, nrn in enumerate(_pick(files)):
        ids = nrn_chk.has_all_nonzero_section_lengths(nrn)
        assert ids.status
        assert len(ids.info) == 0


def test_nonzero_section_lengths_bad_data():
    nrn = NEURONS['Neuron_zero_length_sections.swc']

    ids = nrn_chk.has_all_nonzero_section_lengths(nrn)
    assert not ids.status
    assert ids.info == [13]


def test_nonzero_section_lengths_threshold():
    nrn = NEURONS['Neuron.swc']

    ids = nrn_chk.has_all_nonzero_section_lengths(nrn)
    assert ids.status
    assert len(ids.info) == 0

    ids = nrn_chk.has_all_nonzero_section_lengths(nrn, threshold=15.)
    assert not ids.status
    assert len(ids.info) == 84


def test_has_nonzero_soma_radius():

    nrn = load_neuron(SWC_PATH / 'Neuron.swc')
    assert nrn_chk.has_nonzero_soma_radius(nrn)


def test_has_nonzero_soma_radius_bad_data():
    nrn = load_neuron(SWC_PATH / 'soma_zero_radius.swc')
    assert not nrn_chk.has_nonzero_soma_radius(nrn).status


def test_has_no_fat_ends():
    _, nrn = _load_neuron('fat_end.swc')
    assert not nrn_chk.has_no_fat_ends(nrn).status

    # if we only use point, there isn't a 'fat end'
    # since if the last point is 'x': x < 2*mean([x])
    assert nrn_chk.has_no_fat_ends(nrn, final_point_count=1).status

    # if the multiple of the mean is large, the end won't be fat
    assert nrn_chk.has_no_fat_ends(nrn, multiple_of_mean=10).status

    _, nrn = _load_neuron('Single_basal.swc')
    assert nrn_chk.has_no_fat_ends(nrn).status


def test_has_no_root_node_jumps():
    _, nrn = _load_neuron('root_node_jump.swc')
    check = nrn_chk.has_no_root_node_jumps(nrn)
    assert not check.status
    assert len(check.info) == 1
    assert check.info[0][0] == 0
    assert_array_equal(check.info[0][1], [[0, 3, 0]])

    assert nrn_chk.has_no_root_node_jumps(nrn, radius_multiplier=4).status


def test_has_no_narrow_start():
    _, nrn = _load_neuron('narrow_start.swc')
    check = nrn_chk.has_no_narrow_start(nrn)
    assert not check.status
    assert_array_equal(check.info[0][1][:, COLS.XYZR], [[0, 0, 2, 2]])

    _, nrn = _load_neuron('narrow_start.swc')
    assert nrn_chk.has_no_narrow_start(nrn, 0.25).status

    _, nrn = _load_neuron('fat_end.swc')  # doesn't have narrow start
    assert nrn_chk.has_no_narrow_start(nrn, 0.25).status


def test_has_nonzero_soma_radius_threshold():

    class Dummy:
        pass

    nrn = Dummy()
    nrn.soma = Dummy()
    nrn.soma.radius = 1.5

    assert nrn_chk.has_nonzero_soma_radius(nrn)
    assert nrn_chk.has_nonzero_soma_radius(nrn, 0.25)
    assert nrn_chk.has_nonzero_soma_radius(nrn, 0.75)
    assert nrn_chk.has_nonzero_soma_radius(nrn, 1.25)
    assert nrn_chk.has_nonzero_soma_radius(nrn, 1.499)
    assert not nrn_chk.has_nonzero_soma_radius(nrn, 1.5)
    assert not nrn_chk.has_nonzero_soma_radius(nrn, 1.75)
    assert not nrn_chk.has_nonzero_soma_radius(nrn, 2.5)


def test_has_no_jumps():
    _, nrn = _load_neuron('z_jump.swc')
    assert not nrn_chk.has_no_jumps(nrn).status
    assert nrn_chk.has_no_jumps(nrn, 100).status

    assert nrn_chk.has_no_jumps(nrn, 100, axis='x').status


def test_has_no_narrow_dendritic_section():
    swc_content = StringIO(u"""
# index, type, x, y, z, radius, parent
    1 1  0  0 0 10. -1
    2 2  0  0 0 10.  1
    3 2  0  50 0 10.  2
    4 2 -5  51 0 10.  3
    5 2  6  53 0 10.  3
    6 3  0  0 0 5.  1  # start of the narrow section
    7 3  0 -4 0 5.  6
    8 3  6 -4 0 10.  7
    9 3 -5 -4 0 10.  7
""")
    nrn = load_neuron(swc_content, reader='swc')
    res = nrn_chk.has_no_narrow_neurite_section(nrn,
                                                dendrite_filter,
                                                radius_threshold=5,
                                                considered_section_min_length=0)

    assert res.status

    res = nrn_chk.has_no_narrow_neurite_section(nrn, dendrite_filter,
                                                radius_threshold=7,
                                                considered_section_min_length=0)
    assert not res.status

    swc_content = StringIO(u"""
# index, type, x, y, z, radius, parent
    1 1  0  0 0 10. -1
    2 2  0  0 0 5  1 # narrow soma
    3 2  0  50 0 5  2
    4 2 -5  51 0 5  3
    5 2  6  53 0 5  3
    6 3  0  0 0 5  1 # narrow axon
    7 3  0 -4 0 10.  6
    8 3  6 -4 0 10.  7
    9 3 -5 -4 0 10.  7
""")
    nrn = load_neuron(swc_content, reader='swc')
    res = nrn_chk.has_no_narrow_neurite_section(nrn, dendrite_filter,
                                                radius_threshold=5,
                                                considered_section_min_length=0)
    assert res.status, 'Narrow soma or axons should not raise bad status when checking for narrow dendrites'


def test_has_no_dangling_branch():
    _, nrn = _load_neuron('dangling_axon.swc')
    res = nrn_chk.has_no_dangling_branch(nrn)
    assert not res.status
    assert len(res.info) == 1
    assert_array_equal(res.info[0][1][0][COLS.XYZ],
                       [0., 49.,  0.])

    _, nrn = _load_neuron('dangling_dendrite.swc')
    res = nrn_chk.has_no_dangling_branch(nrn)
    assert not res.status
    assert len(res.info) == 1
    assert_array_equal(res.info[0][1][0][COLS.XYZ],
                       [0., 49.,  0.])

    _, nrn = _load_neuron('axon-sprout-from-dendrite.asc')
    res = nrn_chk.has_no_dangling_branch(nrn)
    assert res.status


def test__bool__():
    c = check.CheckResult(status=True)
    assert c.__nonzero__()
    assert c.__bool__() == c.__nonzero__()



def test_has_multifurcation():
    nrn = load_neuron(StringIO(u"""
	((CellBody) (0 0 0 2))
( (Color Blue)
  (Axon)
  (0 5 0 2)
  (2 9 0 2)
  (0 13 0 2)
  (
    (0 13 0 2)
    (4 13 0 2)
    |
    (0 13 0 2)
    (4 13 0 2)
    |
    (0 13 0 2)
    (4 13 0 2)
    |
    (0 13 0 2)
    (4 13 0 2)
  )
)
"""), reader='asc')

    check_ = nrn_chk.has_multifurcation(nrn)
    assert not check_.status
    info = check_.info
    assert_array_equal(info[0][0], 0)
    assert_array_equal(info[0][1][:, COLS.XYZR], [[0.0, 13.0, 0.0, 1.0]])


def test_single_children():
    neuron = load_neuron("""
( (Color Blue)
  (Axon)
  (0 5 0 2)
  (2 9 0 2)
  (0 13 0 2)
  (
    (2 13 0 2)
    (4 13 0 2)
    (6 13 0 2)
  )
)
""", "asc")
    result = nrn_chk.has_no_single_children(neuron)
    assert result.status == False
    assert result.info == [0]



# Copyright (c) 2020, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""NeuroM, lightweight and fast.

Examples:
    Obtain some morphometrics
    >>> import neurom
    >>> from neurom import features
    >>> nrn = neurom.load_neuron('path/to/neuron')
    >>> ap_seg_len = features.get('segment_lengths', nrn, neurite_type=neurom.APICAL_DENDRITE)
    >>> ax_sec_len = features.get('section_lengths', nrn, neurite_type=neurom.AXON)
"""

import numpy as np

from neurom.core.types import NeuriteType, tree_type_checker
from neurom.core.neuron import iter_neurites
from neurom.exceptions import NeuroMError
from neurom.utils import deprecated

NEURITEFEATURES = dict()
NEURONFEATURES = dict()


@deprecated(
    '`register_neurite_feature`',
    'Please use the decorator `neurom.features.register.feature` to register custom features')
def register_neurite_feature(name, func):
    """Register a feature to be applied to neurites.

    .. warning:: This feature has been deprecated in 1.6.0

    Arguments:
        name: name of the feature, used for access via get() function.
        func: single parameter function of a neurite.

    """
    def _fun(neurites, neurite_type=NeuriteType.all):
        """Wrap neurite function from outer scope and map into list."""
        return list(func(n) for n in iter_neurites(neurites, filt=tree_type_checker(neurite_type)))

    _register_feature('NEURITEFEATURES', name, _fun, shape=(...,))


def _find_feature_func(feature_name):
    """Returns the python function used when getting a feature with `neurom.get(feature_name)`."""
    for feature_dict in (NEURITEFEATURES, NEURONFEATURES):
        if feature_name in feature_dict:
            return feature_dict[feature_name]
    raise NeuroMError(f'Unable to find feature: {feature_name}')


def _get_feature_value_and_func(feature_name, obj, **kwargs):
    """Obtain a feature from a set of morphology objects.

    Arguments:
        feature(string): feature to extract
        obj: a neuron, population or neurite tree
        kwargs: parameters to forward to underlying worker functions

    Returns:
        A tuple (feature, func) of the feature value and its function
    """
    feat = _find_feature_func(feature_name)

    res = feat(obj, **kwargs)
    if len(feat.shape) != 0:
        res = np.array(list(res))

    return res, feat


def get(feature_name, obj, **kwargs):
    """Obtain a feature from a set of morphology objects.

    Arguments:
        feature_name(string): feature to extract
        obj: a neuron, population or neurite tree
        kwargs: parameters to forward to underlying worker functions

    Returns:
        features as a 1D, 2D or 3D numpy array.
    """
    return _get_feature_value_and_func(feature_name, obj, **kwargs)[0]


_INDENT = ' ' * 4


def _indent(string, count):
    """Indent `string` by `count` * INDENT."""
    indent = _INDENT * count
    ret = indent + string.replace('\n', '\n' + indent)
    return ret.rstrip()


def _get_doc():
    """Get a description of all the known available features."""
    def get_docstring(func):
        """Extract doctstring, if possible."""
        docstring = ':\n'
        if func.__doc__:
            docstring += _indent(func.__doc__, 2)
        return docstring

    ret = ['\nNeurite features (neurite, neuron, neuron population):']
    ret.extend(_INDENT + '- ' + feature + get_docstring(func)
               for feature, func in sorted(NEURITEFEATURES.items()))

    ret.append('\nNeuron features (neuron, neuron population):')
    ret.extend(_INDENT + '- ' + feature + get_docstring(func)
               for feature, func in sorted(NEURONFEATURES.items()))

    return '\n'.join(ret)


def _register_feature(namespace, name, func, shape):
    """Register a feature to be applied.

    Upon registration, an attribute 'shape' containing the expected
    shape of the function return is added to 'func'.

    Arguments:
        namespace(string): a namespace (must be 'NEURITEFEATURES' or 'NEURONFEATURES')
        name(string): name of the feature, used to access the feature via `neurom.features.get()`.
        func(callable): single parameter function of a neurite.
        shape(tuple): the expected shape of the feature values
    """
    setattr(func, 'shape', shape)

    assert namespace in {'NEURITEFEATURES', 'NEURONFEATURES'}
    feature_dict = globals()[namespace]

    if name in feature_dict:
        raise NeuroMError('Attempt to hide registered feature %s' % name)
    feature_dict[name] = func


def feature(shape, namespace=None, name=None):
    """Feature decorator to automatically register the feature in the appropriate namespace.

    Arguments:
        shape(tuple): the expected shape of the feature values
        namespace(string): a namespace (must be 'NEURITEFEATURES' or 'NEURONFEATURES')
        name(string): name of the feature, used to access the feature via `neurom.features.get()`.
    """
    def inner(func):
        _register_feature(namespace, name or func.__name__, func, shape)
        return func
    return inner


# These imports are necessary in order to register the features
from neurom.features import neuritefunc, neuronfunc  # noqa, pylint: disable=wrong-import-position

# This must be done after all features have been registered
get.__doc__ += _indent('\nFeatures:\n', 1) + _indent(_get_doc(), 2)  # pylint: disable=no-member



# Copyright (c) 2015, Ecole Polytechnique Federale de Lausanne, Blue Brain Project
# All rights reserved.
#
# This file is part of NeuroM <https://github.com/BlueBrain/NeuroM>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#        notice, this list of conditions and the following disclaimer.
#     2. Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#     3. Neither the name of the copyright holder nor the names of
#        its contributors may be used to endorse or promote products
#        derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Test neurom.io.utils."""
import warnings
from io import StringIO
from pathlib import Path

import numpy as np
from morphio import MissingParentError, RawDataError, SomaError, UnknownFileType, MorphioError
from neurom import COLS, get, load_neuron
from neurom.core.neuron import Neuron
from neurom.exceptions import NeuroMError
from neurom.io import utils
import pytest

DATA_PATH = Path(__file__).parent.parent / 'data'
SWC_PATH = DATA_PATH / 'swc'
VALID_DATA_PATH = DATA_PATH / 'valid_set'
NRN_NAMES = ('Neuron', 'Neuron_h5v1')
FILES = [SWC_PATH / f
         for f in ['Neuron.swc',
                   'Single_apical_no_soma.swc',
                   'Single_apical.swc',
                   'Single_basal.swc',
                   'Single_axon.swc',
                   'sequential_trunk_off_0_16pt.swc',
                   'sequential_trunk_off_1_16pt.swc',
                   'sequential_trunk_off_42_16pt.swc',
                   'Neuron_no_missing_ids_no_zero_segs.swc']]
FILENAMES = [VALID_DATA_PATH / f
             for f in ['Neuron.swc', 'Neuron_h5v1.h5']]
NRN = utils.load_neuron(VALID_DATA_PATH / 'Neuron.swc')
NO_SOMA_FILE = SWC_PATH / 'Single_apical_no_soma.swc'
DISCONNECTED_POINTS_FILE = SWC_PATH / 'Neuron_disconnected_components.swc'
MISSING_PARENTS_FILE = SWC_PATH / 'Neuron_missing_parents.swc'


def _mock_load_neuron(filename):
    class MockNeuron:
        def __init__(self, name):
            self.soma = 42
            self.neurites = list()
            self.name = name

    return MockNeuron(Path(filename).stem)


def _check_neurites_have_no_parent(nrn):

    for n in nrn.neurites:
        assert n.root_node.parent is None


def test_get_morph_files():
    ref = {'Neuron_slice.h5', 'Neuron.swc', 'Neuron_h5v1.h5', 'Neuron_2_branch_h5v1.h5'}

    files = set(f.name for f in utils.get_morph_files(VALID_DATA_PATH))
    assert ref == files


def test_load_neurons():
    # List of strings
    nrns = utils.load_neurons(map(str, FILES), neuron_loader=_mock_load_neuron)
    for i, nrn in enumerate(nrns):
        assert nrn.name == FILES[i].stem

    with pytest.raises(NeuroMError):
        utils.load_neurons(MISSING_PARENTS_FILE,)

    # Single string
    nrns = utils.load_neurons(str(FILES[0]), neuron_loader=_mock_load_neuron)
    assert nrns[0].name == FILES[0].stem

    # Single Path
    nrns = utils.load_neurons(FILES[0], neuron_loader=_mock_load_neuron)
    assert nrns[0].name == FILES[0].stem

    # sequence of strings
    nrns = utils.load_neurons(map(str, FILES), neuron_loader=_mock_load_neuron)
    for i, nrn in enumerate(nrns):
        assert nrn.name == FILES[i].stem

    # sequence of Path objects
    nrns = utils.load_neurons(FILES, neuron_loader=_mock_load_neuron)
    for nrn, file in zip(nrns, FILES):
        assert nrn.name == file.stem

    # string path to a directory
    nrns = utils.load_neurons(str(SWC_PATH), neuron_loader=_mock_load_neuron)
    # is subset so that if new morpho are added to SWC_PATH, the test does not break
    assert {f.stem for f in FILES}.issubset({nrn.name for nrn in nrns})

    # Path path to a directory
    nrns = utils.load_neurons(SWC_PATH, neuron_loader=_mock_load_neuron)
    # is subset so that if new morpho are added to SWC_PATH, the test does not break
    assert {f.stem for f in FILES}.issubset({nrn.name for nrn in nrns})


def test_ignore_exceptions():
    with pytest.raises(NeuroMError):
        utils.load_neurons(MISSING_PARENTS_FILE,)
    pop = utils.load_neurons((MISSING_PARENTS_FILE,), ignored_exceptions=(RawDataError,))
    assert len(pop) == 0


def test_load_neuron():
    nrn = utils.load_neuron(FILENAMES[0])
    assert isinstance(NRN, Neuron)
    assert NRN.name == 'Neuron'
    _check_neurites_have_no_parent(nrn)

    neuron_str = u""" 1 1  0  0 0 1. -1
                      2 3  0  0 0 1.  1
                      3 3  0  5 0 1.  2
                      4 3 -5  5 0 0.  3
                      5 3  6  5 0 0.  3
                      6 2  0  0 0 1.  1
                      7 2  0 -4 0 1.  6
                      8 2  6 -4 0 0.  7
                      9 2 -5 -4 0 0.  7
                     """
    utils.load_neuron(StringIO(neuron_str), reader='swc')


def test_neuron_name():
    for fn, nn in zip(FILENAMES, NRN_NAMES):
        nrn = utils.load_neuron(fn)
        assert nrn.name == nn


def test_load_bifurcating_soma_points_raises_SomaError():
    with pytest.raises(SomaError):
        utils.load_neuron(Path(SWC_PATH, 'soma', 'bifurcating_soma.swc'))


def test_load_neuromorpho_3pt_soma():
    with warnings.catch_warnings(record=True):
        nrn = utils.load_neuron(Path(SWC_PATH, 'soma', 'three_pt_soma.swc'))
    assert len(nrn.neurites) == 4
    assert len(nrn.soma.points) == 3
    assert nrn.soma.radius == 2
    _check_neurites_have_no_parent(nrn)


def test_neurites_have_no_parent():

    _check_neurites_have_no_parent(NRN)


def test_neuron_sections():
    # check no duplicates
    assert len(set(NRN.sections)) == len(list(NRN.sections))


def test_neuron_sections_are_connected():
    # check traversal by counting number of sections un trees
    for nrt in NRN.neurites:
        root_node = nrt.root_node
        assert (sum(1 for _ in root_node.ipreorder()) ==
                        sum(1 for _ in NRN.sections[root_node.id].ipreorder()))


def test_load_neuron_soma_only():

    nrn = utils.load_neuron(Path(DATA_PATH, 'swc', 'Soma_origin.swc'))
    assert len(nrn.neurites) == 0
    assert nrn.name == 'Soma_origin'


def test_load_neuron_disconnected_points_raises():
    with pytest.raises(MorphioError, match='Warning: found a disconnected neurite'):
        load_neuron(DISCONNECTED_POINTS_FILE)


def test_load_neuron_missing_parents_raises():
    with pytest.raises(MissingParentError):
        utils.load_neuron(MISSING_PARENTS_FILE)


def test_load_neurons_directory():
    pop = utils.load_neurons(VALID_DATA_PATH)
    assert len(pop.neurons) == 4
    assert len(pop) == 4
    assert pop.name == 'valid_set'
    for nrn in pop:
        assert isinstance(nrn, Neuron)


def test_load_neurons_directory_name():
    pop = utils.load_neurons(VALID_DATA_PATH, name='test123')
    assert len(pop.neurons) == 4
    assert len(pop) == 4
    assert pop.name == 'test123'
    for nrn in pop:
        assert isinstance(nrn, Neuron)


def test_load_neurons_filenames():
    pop = utils.load_neurons(FILENAMES, name='test123')
    assert len(pop.neurons) == 2
    assert pop.name == 'test123'
    for nrn, name in zip(pop.neurons, NRN_NAMES):
        assert isinstance(nrn, Neuron)
        assert nrn.name == name


SWC_ORD_PATH = Path(DATA_PATH, 'swc', 'ordering')
SWC_ORD_REF = utils.load_neuron(Path(SWC_ORD_PATH, 'sample.swc'))


def assert_items_equal(a, b):
    assert sorted(a) == sorted(b)


def test_load_neuron_mixed_tree_swc():
    nrn_mix = utils.load_neuron(Path(SWC_ORD_PATH, 'sample_mixed_tree_sections.swc'))
    assert_items_equal(get('number_of_sections_per_neurite', nrn_mix), [5, 3])

    assert_items_equal(get('number_of_sections_per_neurite', nrn_mix),
                       get('number_of_sections_per_neurite', SWC_ORD_REF))

    assert_items_equal(get('number_of_segments', nrn_mix),
                       get('number_of_segments', SWC_ORD_REF))

    assert_items_equal(get('total_length', nrn_mix),
                       get('total_length', SWC_ORD_REF))


def test_load_neuron_section_order_break_swc():
    nrn_mix = utils.load_neuron(Path(SWC_ORD_PATH, 'sample_disordered.swc'))

    assert_items_equal(get('number_of_sections_per_neurite', nrn_mix), [5, 3])

    assert_items_equal(get('number_of_sections_per_neurite', nrn_mix),
                       get('number_of_sections_per_neurite', SWC_ORD_REF))

    assert_items_equal(get('number_of_segments', nrn_mix),
                       get('number_of_segments', SWC_ORD_REF))

    assert_items_equal(get('total_length', nrn_mix),
                       get('total_length', SWC_ORD_REF))


H5_PATH = Path(DATA_PATH, 'h5', 'v1', 'ordering')
H5_ORD_REF = utils.load_neuron(Path(H5_PATH, 'sample.h5'))


def test_load_neuron_mixed_tree_h5():
    nrn_mix = utils.load_neuron(Path(H5_PATH, 'sample_mixed_tree_sections.h5'))
    assert_items_equal(get('number_of_sections_per_neurite', nrn_mix), [5, 3])
    assert_items_equal(get('number_of_sections_per_neurite', nrn_mix),
                       get('number_of_sections_per_neurite', H5_ORD_REF))


def test_load_h5_trunk_points_regression():
    # regression test for issue encountered while
    # implementing PR #479, related to H5 unpacking
    # of files with non-standard soma structure.
    # See #480.
    nrn = utils.load_neuron(Path(DATA_PATH, 'h5', 'v1', 'Neuron.h5'))
    assert np.allclose(nrn.neurites[0].root_node.points[1, COLS.XYZR],
                       [0., 0., 0.1, 0.31646374])

    assert np.allclose(nrn.neurites[1].root_node.points[1, COLS.XYZR],
                       [0., 0., 0.1, 1.84130445e-01])

    assert np.allclose(nrn.neurites[2].root_node.points[1, COLS.XYZR],
                       [0., 0., 0.1, 5.62225521e-01])

    assert np.allclose(nrn.neurites[3].root_node.points[1, COLS.XYZR],
                       [0., 0., 0.1, 7.28555262e-01])


def test_load_unknown_type():
    with pytest.raises(UnknownFileType):
        load_neuron(DATA_PATH / 'unsupported_extension.fake')


def test_NeuronLoader():
    dirpath = Path(DATA_PATH, 'h5', 'v1')
    loader = utils.NeuronLoader(dirpath, file_ext='.h5', cache_size=5)
    nrn = loader.get('Neuron')
    assert isinstance(nrn, Neuron)
    # check caching
    assert nrn == loader.get('Neuron')
    assert nrn != loader.get('Neuron_2_branch')


def test_NeuronLoader_mixed_file_extensions():
    loader = utils.NeuronLoader(VALID_DATA_PATH)
    loader.get('Neuron')
    loader.get('Neuron_h5v1')
    with pytest.raises(NeuroMError):
        loader.get('NoSuchNeuron')


def test_get_files_by_path():
    single_neurom = utils.get_files_by_path(NO_SOMA_FILE)
    assert len(single_neurom) == 1

    neuron_dir = utils.get_files_by_path(VALID_DATA_PATH)
    assert len(neuron_dir) == 4

    with pytest.raises(IOError):
        utils.get_files_by_path(Path('this/is/a/fake/path'))

def test_h5v2_raises():
    with pytest.raises(RawDataError):
        utils.load_neuron(DATA_PATH / 'h5/v2/Neuron.h5')
